# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Utilities for unified input/output."""

from __future__ import annotations

import gzip
import os
import tempfile
import warnings
from functools import wraps
from io import IOBase
from typing import (
    TYPE_CHECKING,
    Protocol,
    TypeVar,
    cast,
    overload,
    runtime_checkable,
)
from urllib.parse import urlparse

if TYPE_CHECKING:
    from collections.abc import Callable
    from typing import (
        ParamSpec,
        TypeAlias,
    )

    from lal.utils import CacheEntry

    # Type variables for decorator typing
    P = ParamSpec("P")
    T = TypeVar("T")

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

__all__ = [
    "FILE_LIKE",
    "FileLike",
    "NamedFileLike",
    "NamedReadable",
    "NamedWritable",
    "Readable",
    "Writable",
    "file_list",
    "file_path",
    "gopen",
    "with_open",
]


# -- IO type hints -------------------

AnyStr = TypeVar("AnyStr", str, bytes)

@runtime_checkable
class NamedIO(Protocol[AnyStr]):
    """Typing protocol for file-like objects with a name."""

    name: str

    # Core IO methods from typing.IO[str]
    def close(self) -> None:
        """Close the file."""
    def flush(self) -> None:
        """Flush the file."""
    def isatty(self) -> bool:
        """Return True if the file is connected to a terminal."""
    def read(self, n: int = -1) -> AnyStr:
        """Read n bytes from the file."""
    def readable(self) -> bool:
        """Return True if the file is readable."""
    def readline(self, limit: int = -1) -> AnyStr:
        """Read a single line from the file."""
    def readlines(self, hint: int = -1) -> list[AnyStr]:
        """Read all lines from the file."""
    def seek(self, offset: int, whence: int = 0) -> int:
        """Seek to a position in the file."""
    def seekable(self) -> bool:
        """Return True if the file supports seeking."""
    def tell(self) -> int:
        """Return the current position in the file."""
    def truncate(self, size: int | None = None) -> int:
        """Truncate the file to a given size."""
    def writable(self) -> bool:
        """Return True if the file is writable."""
    def write(self, s: AnyStr) -> int:
        """Write data to the file."""
    def writelines(self, lines: list[AnyStr]) -> None:
        """Write a list of lines to the file."""

    # Context manager
    def __enter__(self) -> NamedIO[AnyStr]:
        """Enter the runtime context related to this object."""
    def __exit__(self, exc_type, exc_val, exc_tb) -> None:  # noqa: ANN001
        """Exit the runtime context related to this object."""

    # Iterator protocol
    def __iter__(self) -> NamedIO[AnyStr]:
        """Return an iterator over the lines of the file."""
    def __next__(self) -> AnyStr:
        """Return the next line from the file."""


#: Type alias for file-like objects
FileLike: TypeAlias = IOBase | gzip.GzipFile | tempfile._TemporaryFileWrapper  # noqa: SLF001

#: Type alias for file-like objects with a ``name`` attribute
NamedFileLike: TypeAlias = NamedIO | gzip.GzipFile | tempfile._TemporaryFileWrapper  # noqa: SLF001

#: Type alias for file system paths (that can be opened)
FileSystemPath: TypeAlias = str | os.PathLike

#: Type alias for readable objects (path or file-like)
Readable: TypeAlias = FileSystemPath | FileLike

#: Type alias for writable objects (path or file-like)
Writable = Readable

#: Type alias for named readable objects (path or named file-like)
NamedReadable: TypeAlias = FileSystemPath | NamedFileLike

#: Type alias for named writable objects (path or named file-like)
NamedWritable = NamedReadable

#: Tuple of file-like types
#:
#: .. deprecated:: 4.0.0
#:
#:     This type tuple is deprecated, use `FileLike` instead
FILE_LIKE = (
    IOBase,
    gzip.GzipFile,
    tempfile._TemporaryFileWrapper,  # noqa: SLF001
)


# -- File open utilities -------------

@overload
def with_open(
    func: Callable[P, T],
    mode: str = "r",
    pos: int = 0,
) -> Callable[P, T]: ...

@overload
def with_open(
    func: None = None,
    mode: str = "r",
    pos: int = 0,
) -> Callable[[Callable[P, T]], Callable[P, T]]: ...

def with_open(
    func: Callable[P, T] | None = None,
    mode: str = "r",
    pos: int = 0,
) -> Callable[P, T] | Callable[[Callable[P, T]], Callable[P, T]]:
    """Decorate a function to ensure the chosen argument is an open file.

    Parameters
    ----------
    func : `callable`
        the function to decorate

    mode : `str`, optional
        the mode with which to open files

    pos : `int`, optional
        which argument to look at

    Examples
    --------
    To ensure that the first argument is an open read-only file, just use
    the decorator without functional parentheses or arguments:

    >>> @with_open
    >>> def my_func(pathorfile, *args, **kwargs)
    >>>     ...

    To ensure that the second argument (position 1) is a file open for writing:

    >>> @with_open(mode="w", pos=1)
    >>> def my_func(stuff, pathorfile, *args, **kwargs)
    >>>     stuff.write_to(pathorfile, *args, **kwargs)
    """
    def _decorator(func: Callable[P, T]) -> Callable[P, T]:
        @wraps(func)
        def wrapped_func(*args: P.args, **kwargs: P.kwargs) -> T:
            # if the relevant positional argument isn't an open
            # file, or something that looks like one, ...
            source = args[pos]
            if not isinstance(source, FileLike):
                source = cast("FileSystemPath", source)
                # open the file, ...
                with open(source, mode=mode) as fobj:  # noqa: PTH123
                    # replace the argument with the open file, ...
                    args = list(args)  # type: ignore[assignment]
                    args[pos] = fobj  # type: ignore[index]
                    # and re-execute the function call
                    return func(*args, **kwargs)
            return func(*args, **kwargs)
        return wrapped_func
    if func:
        return _decorator(func)
    return _decorator


# -- file list utilities -------------

def file_list(
    flist: NamedReadable | list[NamedReadable] | tuple[NamedReadable, ...],
) -> list[str]:
    """Parse a number of possible input types into a list of filepaths.

    Parameters
    ----------
    flist : `file-like` or `list-like` iterable
        The input data container, normally just a single file path, or a list
        of paths, but can generally be any of the following

        - `str` representing a single file path (or comma-separated collection)
        - open `file` or `~gzip.GzipFile` object
        - :class:`~lal.utils.CacheEntry`
        - `str` with ``.cache`` or ``.lcf`` extension
        - simple `list` or `tuple` of `str` paths

    Returns
    -------
    files : `list`
        `list` of `str` file paths

    Raises
    ------
    ValueError
        if the input `flist` cannot be interpreted as any of the above inputs
    """
    # open a cache file and return list of paths
    if (
        isinstance(flist, str)
        and flist.endswith((".cache", ".lcf", ".ffl"))
    ):
        from .cache import read_cache
        return read_cache(flist)

    # separate comma-separate list of names
    if isinstance(flist, str):
        return flist.split(",")

    # parse list of entries (of some format)
    if isinstance(flist, list | tuple):
        return list(map(file_path, flist))

    # otherwise parse a single entry
    try:
        return [file_path(flist)]
    except ValueError as exc:
        exc.args = (
            f"Could not parse input {flist!r} as "
            "one or more file-like objects",
        )
        raise


def file_path(fobj: NamedReadable | bytes | CacheEntry) -> str:
    """Determine the path of a file.

    This doesn't do any sanity checking to check that the file
    actually exists, or is readable.

    Parameters
    ----------
    fobj : `file`, `str`, `os.PathLike`, `bytes`, `CacheEntry`, ...
        The file object or path to parse.

    Returns
    -------
    path : `str`
        The path of the underlying file, always as a `str`.

    Raises
    ------
    ValueError
        If a file path cannnot be determined.

    Examples
    --------
    >>> from gwpy.io.utils import file_path
    >>> import pathlib
    >>> file_path("test.txt")
    'test.txt'
    >>> file_path(pathlib.Path('dir') / 'test.txt')
    'dir/test.txt'
    >>> file_path(open("test.txt", "r"))
    'test.txt'
    >>> file_path("file:///home/user/test.txt")
    '/home/user/test.txt'
    """
    if isinstance(fobj, bytes):
        fobj = fobj.decode("utf-8")
    # file:// URL
    if isinstance(fobj, str) and fobj.startswith("file:"):
        return urlparse(fobj).path
    # Path-like object
    if isinstance(fobj, FileSystemPath):
        return os.fspath(fobj)
    # Named file-like object
    if isinstance(fobj, FileLike) and hasattr(fobj, "name"):
        return fobj.name
    # CacheEntry (or any other object with a .path attribute)
    if hasattr(fobj, "path"):
        return os.fspath(fobj.path)
    # Cannot parse
    msg = f"cannot parse file name for {fobj!r}"
    raise ValueError(msg)


# -- deprecated Gzip support ---------

GZIP_SIGNATURE = b"\x1f\x8b\x08"


def gopen(name, *args, **kwargs):
    """Open a file handling optional gzipping.

    .. deprecated:: 4.0.0

        This function is deprecated and will be removed in a future release.

    If ``name`` ends with ``'.gz'``, or if the GZIP file signature is
    found at the beginning of the file, the file will be opened with
    `gzip.open`, otherwise a regular file will be returned from `open`.

    Parameters
    ----------
    name : `str`, `pathlib.Path`
        path or name of file to open.

    *args, **kwargs
        other arguments to pass to either `open` for regular files, or
        `gzip.open` for gzipped files.

    Returns
    -------
    file : `io.TextIoBase`, `file`, `gzip.GzipFile`
        the open file object
    """
    warnings.warn(
        "gwpy.io.utils.gopen is deprecated and will be removed in a future release.",
        DeprecationWarning,
        stacklevel=2,
    )

    # filename declares gzip
    if str(name).endswith(".gz"):
        return gzip.open(name, *args, **kwargs)

    # open regular file
    fobj = open(name, *args, **kwargs)
    sig = fobj.read(3)
    fobj.seek(0)
    if sig == GZIP_SIGNATURE:  # file signature declares gzip
        fobj.close()  # GzipFile won't close orig file when it closes
        return gzip.open(name, *args, **kwargs)
    return fobj



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2022 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Spectrogram object."""

from __future__ import annotations

import warnings
from typing import (
    TYPE_CHECKING,
    cast,
)

import numpy
from astropy.units import Quantity

from ..frequencyseries import FrequencySeries
from ..frequencyseries._fdcommon import _fdfilter
from ..io.registry import UnifiedReadWriteMethod
from ..timeseries import (
    TimeSeries,
    TimeSeriesList,
)
from ..timeseries.core import _format_time
from ..types import (
    Array2D,
    Series,
)
from ..utils.misc import property_alias
from .connect import (
    SpectrogramRead,
    SpectrogramWrite,
)

if TYPE_CHECKING:
    from collections.abc import Callable
    from typing import (
        ClassVar,
        Literal,
        Self,
    )

    from astropy.units import UnitBase
    from astropy.units.typing import QuantityLike
    from numpy.typing import ArrayLike

    from ..detector import Channel
    from ..frequencyseries import SpectralVariance
    from ..plot import Plot
    from ..signal.filter_design import FilterCompatible
    from ..time import SupportsToGps
    from ..types.sliceutils import SliceLike
    from ..typing import (
        ArrayLike1D,
        UnitLike,
    )

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

__all__ = ["Spectrogram", "SpectrogramList"]


def _ordinal(value: float) -> str:
    """Return the ordinal string for a given integer.

    See https://stackoverflow.com/a/20007730/1307974

    Parameters
    ----------
    value : `float`
        The number to convert to ordinal.

    Examples
    --------
    >>> _ordinal(11)
    '11th'
    >>> _ordinal(102)
    '102nd'
    """
    n = int(str(value)[-1])  # last digit

    # Numbers ending >=4 use 'th' as the ordinal suffix
    th_boundary = 4

    idx = int((n // 10 % 10 != 1) * (n % 10 < th_boundary) * n % 10)
    return f"{value}{'tsnrhtdd'[idx::4]}"


class Spectrogram(Array2D):
    """A 2D array holding a spectrogram of time-frequency data.

    Parameters
    ----------
    value : array-like
        Input data array.

    unit : `~astropy.units.Unit`, optional
        Physical unit of these data.

    epoch : `~gwpy.time.LIGOTimeGPS`, `float`, `str`, optional
        GPS epoch associated with these data,
        any input parsable by `~gwpy.time.to_gps` is fine.

    sample_rate : `float`, `~astropy.units.Quantity`, optional
        The rate of samples per second (Hertz).

    times : `array-like`
        The complete array of GPS times accompanying the data for this series.
        This argument takes precedence over `epoch` and `sample_rate` so should
        be given in place of these if relevant, not alongside.

    f0 : `float`, `~astropy.units.Quantity`, optional
        Starting frequency for these data.

    df : `float`, `~astropy.units.Quantity`, optional
        Frequency resolution for these data.

    frequencies : `array-like`
        The complete array of frequencies indexing the data.
        This argument takes precedence over `f0` and `df` so should
        be given in place of these if relevant, not alongside.

    epoch : `~gwpy.time.LIGOTimeGPS`, `float`, `str`, optional
        GPS epoch associated with these data,
        any input parsable by `~gwpy.time.to_gps` is fine.

    name : `str`, optional
        Descriptive title for this array.

    channel : `~gwpy.detector.Channel`, `str`, optional
        Source data stream for these data.

    dtype : `~numpy.dtype`, optional
        Input data type.

    copy : `bool`, optional, default: `False`
        Choose to copy the input data to new memory.

    subok : `bool`, optional, default: `True`
        Allow passing of sub-classes by the array generator.

    Notes
    -----
    Key methods:

    .. autosummary::

       ~Spectrogram.read
       ~Spectrogram.write
       ~Spectrogram.plot
       ~Spectrogram.zpk
    """

    _metadata_slots: ClassVar[tuple[str, ...]] = (
        *Series._metadata_slots,  # noqa: SLF001
        "y0",
        "dy",
        "yindex",
    )
    _default_xunit: ClassVar[UnitBase] = TimeSeries._default_xunit  # noqa: SLF001
    _default_yunit: ClassVar[UnitBase] = FrequencySeries._default_xunit  # noqa: SLF001
    _rowclass = TimeSeries
    _columnclass = FrequencySeries

    def __new__(
        cls,
        data: ArrayLike,
        unit: UnitLike = None,
        t0: SupportsToGps | None = None,
        dt: Quantity | float | None = None,
        f0: Quantity | float | None = None,
        df: Quantity | float | None = None,
        times: ArrayLike1D | None = None,
        frequencies: ArrayLike1D | None = None,
        name: str | None = None,
        channel: Channel | str | None = None,
        **kwargs,
    ) -> Self:
        """Generate a new Spectrogram."""
        # parse t0 or epoch
        epoch = kwargs.pop("epoch", None)
        if epoch is not None and t0 is not None:
            msg = "give only one of epoch or t0"
            raise ValueError(msg)
        if epoch is None and t0 is not None:
            kwargs["x0"] = _format_time(t0)
        elif epoch is not None:
            kwargs["x0"] = _format_time(epoch)
        # parse sample_rate or dt
        if dt is not None:
            kwargs["dx"] = dt
        # parse times
        if times is not None:
            kwargs["xindex"] = times

        # parse y-axis params
        if f0 is not None:
            kwargs["y0"] = f0
        if df is not None:
            kwargs["dy"] = df
        if frequencies is not None:
            kwargs["yindex"] = frequencies

        # generate Spectrogram
        return super().__new__(
            cls,
            data,
            unit=unit,
            name=name,
            channel=channel,
            **kwargs,
        )

    # -- Spectrogram properties ------

    epoch = property_alias(
        TimeSeries.epoch,  # type: ignore[arg-type]
        "GPS epoch for these data.",
    )
    t0 = property_alias(
        TimeSeries.t0,
        "GPS time of first time bin.",
    )
    dt = property_alias(
        TimeSeries.dt,
        "Time (seconds) between successive bins.",
    )
    span = property_alias(
        TimeSeries.span,
        "GPS [start, stop) span for these data.",
    )
    f0 = property_alias(
        Array2D.y0,  # type: ignore[arg-type]
        "Starting frequency for these data.",
    )
    df = property_alias(
        Array2D.dy,  # type: ignore[arg-type]
        "Frequency spacing for these data.",
    )
    times = property_alias(
        Array2D.xindex,  # type: ignore[arg-type]
        "Series of GPS times for each sample",
    )
    frequencies = property_alias(
        Array2D.yindex,  # type: ignore[arg-type]
        "Series of frequencies for these data.",
    )
    band = property_alias(
        Array2D.yspan,  # type: ignore[arg-type]
        "Frequency band described by these data.",
    )

    # -- Spectrogram i/o -------------

    read = UnifiedReadWriteMethod(SpectrogramRead)
    write = UnifiedReadWriteMethod(SpectrogramWrite)

    # -- Spectrogram methods ---------

    def ratio(
        self,
        operand: FrequencySeries | Quantity | Literal["mean", "median"],
    ) -> Spectrogram:
        """Calculate the ratio of this `Spectrogram` against a reference.

        Parameters
        ----------
        operand : `str`, `FrequencySeries`, `Quantity`
            A `~gwpy.frequencyseries.FrequencySeries` or `~astropy.units.Quantity`
            to weight against, or one of

            ``'mean'``
                Weight against the mean of each spectrum in this Spectrogram.

            ``'median'``
                Weight against the median of each spectrum in this Spectrogram.

        Returns
        -------
        spectrogram : `Spectrogram`
            A new `Spectrogram`.

        Raises
        ------
        ValueError
            If ``operand`` is given as a `str` that isn't supported.
        """
        if isinstance(operand, str):
            if operand == "mean":
                operand = self.mean(axis=0)
            elif operand == "median":
                operand = self.median(axis=0)
            else:
                msg = (
                    f"operand {operand!r} unrecognised, please give a "
                    "Quantity or one of: 'mean', 'median'"
                )
                raise ValueError(msg)
        return self / operand

    def plot(
        self,
        method: str = "pcolormesh",
        figsize: tuple[float, float] = (12, 6),
        xscale: str = "auto-gps",
        **kwargs,
    ) -> Plot:
        """Plot the data for this `Spectrogram`.

        Parameters
        ----------
        method : `str`, optional
            The plotting method to use to render this spectrogram,
            either ``'pcolormesh'`` (default) or ``'imshow'``.

        figsize : `tuple` of `float`, optional
            ``(width, height)`` (inches) of the output figure.

        xscale : `str`, optional
            The X-axis scale.

        kwargs
            All keyword arguments are passed along to underlying
            functions, see below for references.

        Returns
        -------
        plot : `~gwpy.plot.Plot`
            The `Plot` containing the data.

        See Also
        --------
        matplotlib.pyplot.figure
            For documentation of keyword arguments used to create the figure.
        matplotlib.figure.Figure.add_subplot
            For documentation of keyword arguments used to create the axes.
        gwpy.plot.Axes.imshow
        gwpy.plot.Axes.pcolormesh
            For documentation of keyword arguments used in rendering the
            `Spectrogram` data.
        """
        return super().plot(
            method=method,
            figsize=figsize,
            xscale=xscale,
            **kwargs,
        )

    @classmethod
    def from_spectra(
        cls,
        *spectra: FrequencySeries,
        **kwargs,
    ) -> Spectrogram:
        """Build a new `Spectrogram` from a list of spectra.

        Parameters
        ----------
        *spectra : `~gwpy.frequencyseries.FrequencySeries`
            One or more frequency series to stack.

        dt : `float`, `~astropy.units.Quantity`, optional
            Stride between given spectra.

        kwargs
            Other keyword arguments to pass to the constructor.

        Returns
        -------
        Spectrogram
            A new `Spectrogram` from a vertical stacking of the spectra
            The new object takes the metadata from the first given
            `~gwpy.frequencyseries.FrequencySeries` if not given explicitly.

        Notes
        -----
        Each `~gwpy.frequencyseries.FrequencySeries` passed to this
        constructor must be the same length.
        """
        data = numpy.vstack([s.value for s in spectra])
        spec1 = spectra[0]
        if not all(s.f0 == spec1.f0 for s in spectra):
            msg = "cannot stack spectra with different f0"
            raise ValueError(msg)
        if not all(s.df == spec1.df for s in spectra):
            msg = "cannot stack spectra with different df"
            raise ValueError(msg)
        kwargs.setdefault("name", spec1.name)
        kwargs.setdefault("channel", spec1.channel)
        kwargs.setdefault("epoch", spec1.epoch)
        kwargs.setdefault("f0", spec1.f0)
        kwargs.setdefault("df", spec1.df)
        kwargs.setdefault("unit", spec1.unit)
        if (
            "dt" not in kwargs
            and "times" not in kwargs
        ):
            try:
                kwargs["dt"] = spectra[1].epoch.gps - spec1.epoch.gps  # type: ignore[union-attr]
            except (
                AttributeError,
                IndexError,
            ) as exc:
                msg = "cannot determine dt (time-spacing) for Spectrogram from inputs"
                raise ValueError(msg) from exc
        return cls(data, **kwargs)

    def percentile(
        self,
        percentile: float,
    ) -> FrequencySeries:
        """Calculate a given spectral percentile for this `Spectrogram`.

        Parameters
        ----------
        percentile : `float`
            percentile (0 - 100) of the bins to compute

        Returns
        -------
        spectrum : `~gwpy.frequencyseries.FrequencySeries`
            the given percentile `FrequencySeries` calculated from this
            `SpectralVaraicence`
        """
        out = numpy.percentile(self.value, percentile, axis=0)
        ordnl = f"{_ordinal(percentile)} percentile"
        if self.name is not None:
            name = f"{self.name}: {ordnl}"
        else:
            name = ordnl
        return FrequencySeries(
            out,
            epoch=self.epoch,
            channel=self.channel,
            name=name,
            f0=self.f0,
            df=self.df,
            unit=self.unit,
            frequencies=(hasattr(self, "_frequencies") and self.frequencies) or None,
        )

    def zpk(
        self,
        zeros: ArrayLike1D,
        poles: ArrayLike1D,
        gain: float,
        *,
        analog: bool = False,
        sample_rate: QuantityLike | None = None,
        unit: str = "rad/s",
        normalize_gain: bool = False,
    ) -> Self:
        """Filter this `Spectrogram` by applying a zero-pole-gain filter.

        Parameters
        ----------
        zeros : `array-like`
            List of zero frequencies (in Hertz).

        poles : `array-like`
            List of pole frequencies (in Hertz).

        gain : `float`
            DC gain of filter.

        analog : `bool`, optional
            Type of ZPK being applied, if `analog=True` all parameters
            will be converted in the Z-domain for digital filtering.

        sample_rate : `float`, `~astropy.units.Quantity`, optional
            Sample rate of data (in Hertz), used to apply a digital filter.
            Defaults to the last frequency value of this `Spectrogram`
            (i.e. the Nyquist frequency).

        unit : `str`, optional
            For analogue ZPK filters, the units in which the zeros and poles are
            specified. Either ``'Hz'`` or ``'rad/s'`` (default).

        normalize_gain : `bool`, optional
            Whether to normalize the gain when converting from Hz to rad/s.

            - `False` (default):
              Multiply zeros/poles by -2π but leave gain unchanged.
              This matches the LIGO GDS **'f' plane** convention
              (``plane='f'`` in ``s2z()``).

            - `True`:
              Normalize gain to preserve frequency response magnitude.
              Gain is scaled by :math:`|∏p_i/∏z_i| · (2π)^{(n_p - n_z)}`.
              Use this when your filter was designed with the transfer
              function :math:`H(f) = k·∏(f-z_i)/∏(f-p_i)` in Hz.
              This matches the LIGO GDS **'n' plane** convention
              (``plane='n'`` in ``s2z()``).

            Only used for analogue filters in Hz (``analog=True, unit="Hz"``).

        Returns
        -------
        specgram : `Spectrogram`
            The frequency-domain filtered version of the input data.

        See Also
        --------
        Spectrogram.filter
            For details on how a digital ZPK-format filter is applied.

        Examples
        --------
        To apply a zpk filter with file poles at 100 Hz, and five zeros at
        1 Hz (giving an overall DC gain of 1e-10)::

            >>> data2 = data.zpk([100]*5, [1]*5, 1e-10)
        """
        return self.filter(
            (zeros, poles, gain),
            analog=analog,
            unit=unit,
            sample_rate=sample_rate,
            normalize_gain=normalize_gain,
        )

    def filter(
        self,
        filt: FilterCompatible,
        *,
        analog: bool = False,
        sample_rate: QuantityLike | None = None,
        unit: str = "rad/s",
        normalize_gain: bool = False,
        inplace: bool = False,
        **kwargs,
    ) -> Self:
        """Apply the given filter to this `Spectrogram`.

        Parameters
        ----------
        filt : `numpy.ndarray` or `tuple`
            The filter to be applied.
            This can be specified in any of the following forms, with
            the appropriate number of elements in the tuple:

            - `numpy.ndarray` - 1D array of FIR filter coefficients.
            - `tuple[numpy.ndarray, numpy.ndarray]` - numerator/demoinator
              polynomials of the transfer function.
            - `numpy.ndarray` - 2D array of SOS coefficients.
            - `tuple[numpy.ndarray, numpy.ndarray, float]` - zero-pole-gain
              representation.

        analog : `bool`, optional
            If `True`, filter definition will be converted from Hertz
            to Z-domain digital representation, default: `False`.

        sample_rate : `float`, `~astropy.units.Quantity`, optional
            Sample rate of data (in Hertz), used to apply a digital filter.
            Defaults to the last frequency value of this `Spectrogram`
            (i.e. the Nyquist frequency).

        unit : `str`, optional
            For analogue ZPK filters, the units in which the zeros and poles are
            specified. Either ``'Hz'`` or ``'rad/s'`` (default).

        normalize_gain : `bool`, optional
            Whether to normalize the gain when converting from Hz to rad/s.

            - `False` (default):
              Multiply zeros/poles by -2π but leave gain unchanged.
              This matches the LIGO GDS **'f' plane** convention
              (``plane='f'`` in ``s2z()``).

            - `True`:
              Normalize gain to preserve frequency response magnitude.
              Gain is scaled by :math:`|∏p_i/∏z_i| · (2π)^{(n_p - n_z)}`.
              Use this when your filter was designed with the transfer
              function :math:`H(f) = k·∏(f-z_i)/∏(f-p_i)` in Hz.
              This matches the LIGO GDS **'n' plane** convention
              (``plane='n'`` in ``s2z()``).

            Only used for analogue filters in Hz (``analog=True, unit="Hz"``).

        inplace : `bool`, optional
            If `True`, this array will be overwritten with the filtered
            version, default: `False`.

        kwargs
            Additional keyword arguments passed to the filter function.

        Returns
        -------
        result : `Spectrogram`
            The filtered version of the input `Spectrogram`,
            if ``inplace=True`` was given, this is just a reference to
            the modified input array.

        Raises
        ------
        ValueError
            If ``filt`` arguments cannot be interpreted properly.
        """
        return cast("Self", _fdfilter(
            self,
            filt,
            analog=analog,
            sample_rate=sample_rate,
            unit=unit,
            normalize_gain=normalize_gain,
            inplace=inplace,
            **kwargs,
        ))

    def variance(
        self,
        bins: ArrayLike1D | None = None,
        low: float | None = None,
        high: float | None = None,
        nbins: int = 500,
        *,
        log: bool = False,
        norm: bool = False,
        density: bool = False,
    ) -> SpectralVariance:
        """Calculate the `SpectralVariance` of this `Spectrogram`.

        Parameters
        ----------
        bins : `~numpy.ndarray`, optional, default `None`
            array of histogram bin edges, including the rightmost edge
        low : `float`, optional, default: `None`
            left edge of lowest amplitude bin, only read
            if ``bins`` is not given
        high : `float`, optional, default: `None`
            right edge of highest amplitude bin, only read
            if ``bins`` is not given
        nbins : `int`, optional, default: `500`
            number of bins to generate, only read if ``bins`` is not
            given
        log : `bool`, optional, default: `False`
            calculate amplitude bins over a logarithmic scale, only
            read if ``bins`` is not given
        norm : `bool`, optional, default: `False`
            normalise bin counts to a unit sum
        density : `bool`, optional, default: `False`
            normalise bin counts to a unit integral

        Returns
        -------
        specvar : `SpectralVariance`
            2D-array of spectral frequency-amplitude counts

        See Also
        --------
        numpy.histogram
            for details on specifying bins and weights
        """
        from ..frequencyseries import SpectralVariance

        return SpectralVariance.from_spectrogram(
            self,
            bins=bins,
            low=low,
            high=high,
            nbins=nbins,
            log=log,
            norm=norm,
            density=density,
        )

    # -- Spectrogram connectors ------

    def crop_frequencies(
        self,
        low: float | Quantity | None = None,
        high: float | Quantity | None = None,
        *,
        copy: bool = False,
    ) -> Spectrogram:
        """Crop this `Spectrogram` to the specified frequencies.

        Parameters
        ----------
        low : `float`, optional
            Lower frequency bound for cropped `Spectrogram`.

        high : `float`, optional
            Upper frequency bound for cropped `Spectrogram`.

        copy : `bool`, optional
            If `False` return a view of the original data,
            otherwise create a fresh memory copy.

        Returns
        -------
        spec : `Spectrogram`
            A new `Spectrogram` with a subset of data from the frequency
            axis
        """
        # Convert floats to Quantities
        if low is not None:
            low = Quantity(low, self._default_yunit)
        if high is not None:
            high = Quantity(high, self._default_yunit)

        # Cast for type checker
        low = cast("Quantity | None", low)
        high = cast("Quantity | None", high)

        # Check low frequency
        if low is not None and low == self.f0:
            low = None
        elif low is not None and low < self.f0:
            warnings.warn(
                "Spectrogram.crop_frequencies given low frequency "
                "cutoff below f0 of the input Spectrogram. Low "
                "frequency crop will have no effect.",
                stacklevel=2,
            )

        # Check high frequency
        peak = Quantity(self.band[1], self.yunit)
        if high is not None and high == peak:
            high = None
        elif high is not None and high > peak:
            warnings.warn(
                "Spectrogram.crop_frequencies given high frequency "
                "cutoff above cutoff of the input Spectrogram. High "
                "frequency crop will have no effect.",
                stacklevel=2,
            )

        # Find low index
        if low is None:
            idx0 = None
        else:
            idx0 = int(float(low.value - self.f0.value) // self.df.value)
        # find high index
        if high is None:
            idx1 = None
        else:
            idx1 = int(float(high.value - self.f0.value) // self.df.value)

        # Crop
        new = self[:, idx0:idx1]
        if copy:
            return new.copy()
        return new

    # -- Spectrogram ufuncs ----------

    def _wrap_function(
        self,
        function: Callable,
        *args,  # noqa: ANN002
        **kwargs,
    ) -> Self | FrequencySeries | TimeSeries | Quantity:
        """Wrap a numpy function."""
        out = super()._wrap_function(function, *args, **kwargs)

        # requested frequency axis, return a FrequencySeries
        if out.ndim == 1 and out.x0.unit == self.y0.unit:
            return self._columnclass(
                out.value,
                name=out.name,
                unit=out.unit,
                epoch=out.epoch,
                channel=out.channel,
                f0=out.x0.value,
                df=out.dx.value,
            )

        # requested time axis, return a TimeSeries
        if out.ndim == 1:
            return self._rowclass(
                out.value,
                name=out.name,
                unit=out.unit,
                epoch=out.epoch,
                channel=out.channel,
                dx=out.dx,
            )

        # otherwise return whatever we got back from super (Quantity)
        return out

    _wrap_function.__doc__ = Array2D._wrap_function.__doc__  # noqa: SLF001

    # -- other -----------------------

    def __getitem__(
        self,
        item: SliceLike | tuple[SliceLike, ...],
    ) -> Self | FrequencySeries:
        """Return a slice of this spectrogram."""
        out = super().__getitem__(item)

        # set epoch manually, because Spectrogram doesn't store self._epoch
        if isinstance(out, self._columnclass):
            out.epoch = self.epoch

        return out


class SpectrogramList(TimeSeriesList):
    """Fancy list representing a list of `Spectrogram`.

    The `SpectrogramList` provides an easy way to collect and organise
    `Spectrogram` for a single `Channel` over multiple segments.

    Parameters
    ----------
    items
        Any number of `Spectrogram` series.

    Returns
    -------
    list
        A new `SpectrogramList`.

    Raises
    ------
    TypeError
        If any elements are not of type `Spectrogram`.
    """

    EntryClass: ClassVar[type[Spectrogram]] = Spectrogram  # type: ignore[assignment]



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Unit test for timeseries module."""

from __future__ import annotations

import operator
from functools import reduce
from io import BytesIO
from typing import (
    TYPE_CHECKING,
    Generic,
    TypeVar,
)

import numpy
import pytest
from astropy import units
from matplotlib import rc_context
from numpy import shares_memory

from ...detector import Channel
from ...segments import Segment, SegmentList
from ...testing import mocks, utils
from ...time import Time
from ...types.tests.test_series import TestSeries as _TestSeries
from .. import TimeSeriesBase, TimeSeriesBaseDict, TimeSeriesBaseList

if TYPE_CHECKING:
    from numpy.typing import DTypeLike

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

GPS_EPOCH = Time(0, format="gps", scale="utc")
RNG = numpy.random.default_rng(seed=1)

TimeSeriesBaseType = TypeVar("TimeSeriesBaseType", bound=TimeSeriesBase)
TimeSeriesBaseDictType = TypeVar(
    "TimeSeriesBaseDictType",
    bound=TimeSeriesBaseDict[TimeSeriesBase],
)
TimeSeriesBaseListType = TypeVar("TimeSeriesBaseListType", bound=TimeSeriesBaseList)
EntryType = TypeVar("EntryType", bound=TimeSeriesBase)


# -- TimeSeriesBase ------------------

class TestTimeSeriesBase(_TestSeries[TimeSeriesBaseType], Generic[TimeSeriesBaseType]):
    """Test `TimeSeriesBase`."""

    TEST_CLASS: type[TimeSeriesBaseType] = TimeSeriesBase

    def test_new(self):
        """Test `gwpy.timeseries.TimeSeriesBase` constructor."""
        array = self.create()
        super().test_new()

        # check time-domain metadata
        assert array.epoch == GPS_EPOCH
        assert array.sample_rate == units.Quantity(1, "Hertz")
        assert array.dt == units.Quantity(1, "second")

    def test_new_epoch_t0(self):
        """Test `gwpy.timeseries.TimeSeriesBase` handling of epoch vs t0."""
        a = self.create(epoch=10)
        b = self.create(t0=10)
        utils.assert_quantity_sub_equal(a, b)
        with pytest.raises(
            ValueError,
            match=r"^give only one of epoch or t0$",
        ):
            self.TEST_CLASS(self.data, epoch=1, t0=1)

    def test_new_sample_rate_dt(self):
        """Test `gwpy.timeseries.TimeSeriesBase` handling of sample_rate vs dt."""
        # check handling of sample_rate vs dt
        a = self.create(sample_rate=100)
        b = self.create(dt=0.01)
        utils.assert_quantity_sub_equal(a, b)
        with pytest.raises(
            ValueError,
            match=r"^give only one of sample_rate or dt$",
        ):
            self.TEST_CLASS(self.data, sample_rate=1, dt=1)

    def test_epoch(self):  # type: ignore[override]
        """Test `gwpy.timeseries.TimeSeriesBase.epoch`."""
        # check basic conversion from t0 -> epoch
        a = self.create(t0=1126259462)
        assert a.epoch == Time("2015-09-14 09:50:45", format="iso")

        # test that we can't delete epoch
        with pytest.raises(AttributeError):
            del a.epoch

        # check None gets preserved
        a.epoch = None
        with pytest.raises(AttributeError):
            a._t0  # noqa: B018

        # check other types
        a.epoch = Time("2015-09-14 09:50:45", format="iso")
        utils.assert_quantity_almost_equal(
            a.t0,
            units.Quantity(1126259462, "s"),
        )

    def test_sample_rate(self):
        """Test `gwpy.timeseries.TimeSeriesBase.sample_rate`."""
        # check basic conversion from dt -> sample_rate
        a = self.create(dt=0.5)
        assert a.sample_rate == 2 * units.Hz

    def test_sample_rate_del(self, array):
        """Test that `sample_rate` cannot be deleted."""
        # test that we can't delete sample_rate
        with pytest.raises(
            AttributeError,
            match=r"(can't delete attribute|has no deleter)",
        ):
            del array.sample_rate

    def test_sample_rate_none(self, array):
        """Test that `sample_rate = None` is effectively a deletion."""
        # check None gets preserved
        array.sample_rate = None
        with pytest.raises(AttributeError, match="_t0"):
            array._t0  # noqa: B018

    @pytest.mark.parametrize(("samp", "dt"), [
        (128 * units.Hz, units.s / 128.),
        (16.384 * units.kiloHertz, units.s / 16384),
        (10 / units.s, units.s / 10),
    ])
    def test_sample_rate_type(self, array, samp, dt):
        """Test that units and types are handled when setting `sample_rate`."""
        array.sample_rate = samp
        utils.assert_quantity_equal(array.dt, dt)

    def test_sample_rate_ghz(self, array):
        """Test that very large sample rates don't get rounded to dt=0.

        Regression: https://gitlab.com/gwpy/gwpy/-/issues/1646
        """
        array.sample_rate = 1e9
        assert array.dt.value > 0.

    def test_duration(self, array):
        """Test `TimeSeriesBase.duration`."""
        assert array.duration == array.t0 + array.shape[0] * array.dt

    # -- test i/o ---------------------

    def test_read_write_csv(self, array: TimeSeriesBaseType):  # noqa: ARG002
        """Test reading and writing CSV files."""
        pytest.skip(f"not implemented for {self.TEST_CLASS.__name__}")

    # -- test methods -----------------

    def test_plot(self, array):
        """Test `TimeSeriesBase.plot`."""
        with rc_context(rc={"text.usetex": False}):
            plot = array.plot()
            line = plot.gca().lines[0]
            utils.assert_array_equal(line.get_xdata(), array.xindex.value)
            utils.assert_array_equal(line.get_ydata(), array.value)
            plot.save(BytesIO(), format="png")
            plot.close()

    @pytest.mark.requires("arrakis")
    @pytest.mark.parametrize("copy", [False, True])
    @pytest.mark.parametrize("dtype", [
        pytest.param(numpy.float64, id="float64"),
        pytest.param(numpy.int32, id="int32"),
        pytest.param(bool, id="bool"),
        pytest.param(numpy.dtype('bool'), id="np_bool"),
    ])
    def test_from_arrakis(self, copy, dtype):
        """Test `TimeSeriesBase.from_arrakis`."""
        from arrakis import Channel as ArrakisChannel
        from arrakis.block import Series as ArrakisSeries

        data = self.data.astype(dtype)

        # create arrakis objects
        achan = ArrakisChannel(
            "X1:TEST-CHANNEL",
            dtype,
            128,
        )
        aseries = ArrakisSeries(
            data=data,
            time_ns=int(1e18),
            channel=achan,
        )

        # convert to TimeSeries
        ts = self.TEST_CLASS.from_arrakis(aseries, copy=copy)

        # check data
        utils.assert_array_equal(ts.value, data)
        assert shares_memory(ts.value, aseries.data) is not copy

        # check metadata
        assert ts.name == aseries.name
        assert ts.t0 == aseries.t0 * units.s
        assert ts.dt == aseries.dt * units.s
        assert ts.sample_rate == aseries.sample_rate * units.Hz
        assert ts.dtype == numpy.dtype(dtype)
        assert abs(ts.span) == aseries.duration
        utils.assert_array_equal(ts.times.value, aseries.times)
        # see https://git.ligo.org/ngdd/arrakis-python/-/issues/22:
        assert ts.unit == units.dimensionless_unscaled

    @pytest.mark.requires("nds2")
    def test_from_nds2_buffer(self):
        """Test `TimeSeriesBase.from_nds2_buffer`."""
        # build fake buffer
        nds_buffer = mocks.nds2_buffer(
            "X1:TEST",
            self.data,
            1000000000,
            self.data.shape[0],
            "m",
            name="test",
            slope=2,
            offset=1,
        )

        # convert to TimeSeries
        a = self.TEST_CLASS.from_nds2_buffer(nds_buffer)

        # check everything works (including default dynamic scaling)
        assert isinstance(a, self.TEST_CLASS)
        assert not shares_memory(a.value, nds_buffer.data)
        utils.assert_array_equal(a.value, nds_buffer.data * 2 + 1)
        assert a.t0 == 1000000000 * units.s
        assert a.dt == units.s / nds_buffer.data.shape[0]
        assert a.name == "test"
        assert a.channel == Channel(
            "X1:TEST",
            sample_rate=self.data.shape[0],
            unit="m",
            type="raw",
            dtype="float32",
        )

        # check that we can use keywords to override settings
        b = self.TEST_CLASS.from_nds2_buffer(
            nds_buffer,
            scaled=False,
            copy=False,
            sample_rate=128,
        )
        assert b.dt == 1/128. * units.s
        assert shares_memory(nds_buffer.data, b.value)

    @pytest.mark.requires("lal")
    def test_to_from_lal(self, array):
        """Test `TimeSeriesBase.to_lal` and `TimeSeriesBase.from_lal`."""
        # check that to + from returns the same array
        lalts = array.to_lal()
        a2 = type(array).from_lal(lalts)
        utils.assert_quantity_sub_equal(array, a2, exclude=["channel"])

    @pytest.mark.requires("lal")
    @pytest.mark.parametrize("copy", [False, True])
    def test_to_from_lal_no_copy(self, array, copy):
        """Check that copy=False shares data."""
        lalts = array.to_lal()
        a2 = type(array).from_lal(lalts, copy=copy)
        assert shares_memory(a2.value, lalts.data.data) is not copy

    @pytest.mark.requires("lal")
    def test_to_from_lal_unrecognised_units(self, array):
        """Test that unrecognised units get warned, but the operation continues."""
        import lal
        array.override_unit("undef")
        with pytest.warns(
            UserWarning,
            match="defaulting to lal.DimensionlessUnit",
        ):
            lalts = array.to_lal()
        assert lalts.sampleUnits == lal.DimensionlessUnit
        a2 = self.TEST_CLASS.from_lal(lalts)
        assert a2.unit == units.dimensionless_unscaled

    def test_to_from_lal_pow10_units(self, array):
        """Test that normal scaled units scale the data properly."""
        lal = pytest.importorskip("lal")
        array.override_unit("km")
        lalts = array.to_lal()
        utils.assert_array_equal(lalts.data.data, array.value)
        assert lalts.sampleUnits == lal.MeterUnit * 1000.

    def test_to_from_lal_scaled_units(self, array):
        """Test that weird scaled units scale the data properly."""
        lal = pytest.importorskip("lal")
        array.override_unit("123 m")
        lalts = array.to_lal()
        utils.assert_array_equal(lalts.data.data, array.value * 123.)
        assert lalts.sampleUnits == lal.MeterUnit

    @pytest.mark.requires("lal", "pycbc")
    def test_to_from_pycbc(self, array):
        """Test `TimeSeriesBase.to_pycbc` and `TimeSeriesBase.from_pycbc`."""
        from pycbc.types import TimeSeries as PyCBCTimeSeries

        # test default conversion
        pycbcts = array.to_pycbc()
        assert isinstance(pycbcts, PyCBCTimeSeries)
        utils.assert_array_equal(array.value, pycbcts.data)
        assert array.t0.value == pycbcts.start_time
        assert array.dt.value == pycbcts.delta_t

        # go back and check we get back what we put in in the first place
        a2 = type(array).from_pycbc(pycbcts)
        utils.assert_quantity_sub_equal(
            array, a2, exclude=["name", "unit", "channel"])

        # test copy=False
        a2 = type(array).from_pycbc(array.to_pycbc(copy=False), copy=False)
        assert shares_memory(array.value, a2.value)


# -- TimeSeriesBaseDict --------------

class TestTimeSeriesBaseDict(Generic[TimeSeriesBaseDictType, EntryType]):
    """Test base class for TimeSeriesBaseDict."""

    TEST_CLASS: type[TimeSeriesBaseDictType] = TimeSeriesBaseDict
    ENTRY_CLASS: type[EntryType] = TimeSeriesBase
    DTYPE: DTypeLike = None

    @classmethod
    def create(cls) -> TimeSeriesBaseDictType:
        """Create a new `TimeSeriesBaseDict` instance."""
        new = cls.TEST_CLASS()
        new["a"] = cls.ENTRY_CLASS(
            RNG.normal(size=200),
            name="a",
            x0=0,
            dx=1,
            dtype=cls.DTYPE,
        )
        new["b"] = cls.ENTRY_CLASS(
            RNG.normal(size=2000),
            name="b",
            x0=0,
            dx=.1,
            dtype=cls.DTYPE,
        )
        return new

    @pytest.fixture
    def instance(self) -> TimeSeriesBaseDictType:
        """Create an instance of a `TimeSeriesBaseDict`."""
        return self.create()

    def test_series_link(self):
        """Test the links between `DictClass` and `EntryClass`."""
        assert self.ENTRY_CLASS.DictClass is self.TEST_CLASS
        assert self.TEST_CLASS.EntryClass is self.ENTRY_CLASS

    def test_span(self, instance):
        """Test `TimeSeriesBaseDict.span`."""
        assert isinstance(instance.span, Segment)
        assert instance.span == reduce(
            operator.or_, (ts.span for ts in instance.values()), Segment(0, 0),
        )

    def test_span_error_empty(self):
        """Test that `TimeSeriesBaseDict.span` raises for empty dict."""
        with pytest.raises(
            ValueError,
            match="cannot calculate span for empty ",
        ):
            self.TEST_CLASS().span  # noqa: B018

    def test_copy(self, instance):
        """Test `TimeSeriesBaseDict.copy`."""
        copy = instance.copy()
        assert isinstance(copy, self.TEST_CLASS)
        for key in copy:
            assert not shares_memory(copy[key].value, instance[key].value)
            utils.assert_quantity_sub_equal(copy[key], instance[key])

    def test_append(self, instance):
        """Test `TimeSeriesBaseDict.append`."""
        # test appending from empty (with and without copy)
        for copy in (True, False):
            new = type(instance)()
            new.append(instance, copy=copy)
            for key in new:
                assert shares_memory(
                    new[key].value,
                    instance[key].value,
                ) is not copy
                utils.assert_quantity_sub_equal(new[key], instance[key])

        # create copy of dict that is contiguous
        new = type(instance)()
        for key in instance:
            a = instance[key]
            new[key] = type(a)(
                [1, 2, 3, 4, 5],
                x0=a.xspan[1],
                dx=a.dx,
                dtype=a.dtype,
            )

        # append and test
        b = instance.copy()
        b.append(new)
        for key in b:
            utils.assert_array_equal(
                b[key].value,
                numpy.concatenate((instance[key].value, new[key].value)),
            )

        # create copy of dict that is discontiguous
        new = type(instance)()
        for key in instance:
            a = instance[key]
            new[key] = type(a)(
                [1, 2, 3, 4, 5],
                x0=a.xspan[1]+a.dx.value,
                dx=a.dx,
                dtype=a.dtype,
            )
        # check error
        with pytest.raises(
            ValueError,
            match="Cannot append discontiguous",
        ):
            instance.append(new)

        # check padding works (don't validate too much, that is tested
        # elsewhere)
        b = instance.copy()
        b.append(new, pad=0, gap="pad")

    def test_prepend(self, instance):
        """Test `TimeSeriesBaseDict.prepend`."""
        # test appending from empty (with and without copy)
        new = type(instance)()
        new.prepend(instance)
        for key in new:
            assert shares_memory(new[key].value, instance[key].value)
            utils.assert_quantity_sub_equal(new[key], instance[key])

        # create copy of dict that is contiguous
        new = type(instance)()
        for key in instance:
            a = instance[key]
            new[key] = type(a)(
                [1, 2, 3, 4, 5],
                x0=a.xspan[1],
                dx=a.dx,
                dtype=a.dtype,
            )
        # append and test
        b = new.copy()
        b.prepend(instance)
        for key in b:
            utils.assert_array_equal(
                b[key].value,
                numpy.concatenate((instance[key].value, new[key].value)),
            )

        # create copy of dict that is discontiguous
        new = type(instance)()
        for key in instance:
            a = instance[key]
            new[key] = type(a)(
                [1, 2, 3, 4, 5],
                x0=a.xspan[1],
                dx=a.dx,
                dtype=a.dtype,
            )
        # check error
        with pytest.raises(
            ValueError,
            match="Cannot append discontiguous",
        ):
            new.append(instance)
        # check padding works (don't validate too much, that is tested
        # elsewhere)
        b = new.copy()
        b.prepend(instance, pad=0)

    def test_crop(self, instance):
        """Test `TimeSeriesBaseDict.crop`."""
        a = instance.copy().crop(10, 20)  # crop() modifies in-place
        for key in a:
            utils.assert_quantity_sub_equal(a[key], instance[key].crop(10, 20))

    def test_resample(self, instance):
        """Test `TimeSeriesBaseDict.resample`."""
        if self.ENTRY_CLASS is TimeSeriesBase:  # currently only for subclasses
            pytest.skip(f"not implemented for {type(instance).__name__}")

        # for all subclasses
        a = instance.resample(.5)
        for key in a:
            assert a[key].dx == 1/.5 * a[key].xunit

    @pytest.mark.requires("arrakis")
    @pytest.mark.parametrize("copy", [False, True])
    def test_from_arrakis(self, copy):
        """Test :meth:`TimeSeriesBaseDict.from_arrakis`."""
        from arrakis import (
            Channel as ArrakisChannel,
            SeriesBlock as ArrakisBlock,
        )

        # arrakis metadata (channels)
        channels = {
            "X1:TEST-CHANNEL_1": ArrakisChannel(
                "X1:TEST-CHANNEL_1",
                data_type=numpy.dtype(self.DTYPE),
                sample_rate=64,
            ),
            "X1:TEST-CHANNEL_2": ArrakisChannel(
                "X1:TEST-CHANNEL_2",
                data_type=numpy.dtype(self.DTYPE),
                sample_rate=128,
            ),
        }

        # arrakis block
        block = ArrakisBlock(
            int(1e18),
            {
                chan.name: RNG.random(
                    size=int(10 * chan.sample_rate),
                ).astype(self.DTYPE)
                for chan in channels.values()
            },
            channels,
        )

        # convert to TimeSeries
        tsd = self.TEST_CLASS.from_arrakis(block, copy=copy)

        # check keys match
        assert list(tsd) == list(map(str, block.channels))

        for key in tsd:
            # check data
            aseries = block[str(key)]
            ts = tsd[key]
            utils.assert_array_equal(ts.value, aseries.data)
            assert shares_memory(ts.value, aseries.data) is not copy

            # check metadata
            assert ts.name == aseries.name
            assert ts.t0 == aseries.t0 * units.s
            assert ts.dt == aseries.dt * units.s
            assert ts.sample_rate == aseries.sample_rate * units.Hz
            assert abs(ts.span) == aseries.duration
            utils.assert_array_equal(ts.times.value, aseries.times)
            # see https://git.ligo.org/ngdd/arrakis-python/-/issues/22:
            assert ts.unit == units.dimensionless_unscaled

    @pytest.mark.requires("nds2")
    def test_from_nds2_buffers(self):
        """Test `TimeSeriesBaseDict.from_nds2_buffers`."""
        buffers = [
            mocks.nds2_buffer("X1:TEST", numpy.arange(100), 1000000000,
                              1, "m"),
            mocks.nds2_buffer("X1:TEST2", numpy.arange(100, 200), 1000000100,
                              1, "m"),
        ]
        a = self.TEST_CLASS.from_nds2_buffers(buffers)
        assert isinstance(a, self.TEST_CLASS)
        assert a["X1:TEST"].x0.value == 1000000000
        assert a["X1:TEST2"].dx.value == 1
        assert a["X1:TEST2"].x0.value == 1000000100

        a = self.TEST_CLASS.from_nds2_buffers(buffers, sample_rate=.01)
        assert a["X1:TEST"].dx.value == 100

    def test_plot(self, instance):
        """Test `TimeSeriesBaseDict.plot`."""
        with rc_context(rc={"text.usetex": False}):
            plot = instance.plot()
            for line, key in zip(plot.gca().lines, instance, strict=True):
                utils.assert_array_equal(line.get_xdata(),
                                         instance[key].xindex.value)
                utils.assert_array_equal(line.get_ydata(),
                                         instance[key].value)
            plot.save(BytesIO(), format="png")
            plot.close()

    def test_plot_separate(self, instance):
        """Test plotting `TimeSeriesDict` on separate axes.

        See https://gitlab.com/gwpy/gwpy/-/issues/1609
        """
        with rc_context(rc={"text.usetex": False}):
            plot = instance.plot(separate=True)
            assert len(plot.axes) == len(instance.keys())
            for ax, key in zip(plot.axes, instance, strict=True):
                utils.assert_array_equal(ax.lines[-1].get_xdata(),
                                         instance[key].xindex.value)
                utils.assert_array_equal(ax.lines[-1].get_ydata(),
                                         instance[key].value)
            plot.save(BytesIO(), format="png")
            plot.close()


# -- TimeSeriesBaseList --------------

class TestTimeSeriesBaseList(Generic[TimeSeriesBaseListType, EntryType]):
    """Test base class for TimeSeriesBaseList."""

    TEST_CLASS: type[TimeSeriesBaseListType] = TimeSeriesBaseList
    ENTRY_CLASS: type[EntryType] = TimeSeriesBase
    DTYPE: DTypeLike = None

    @classmethod
    def create(cls) -> TimeSeriesBaseListType:
        """Create a new `TimeSeriesList`."""
        new = cls.TEST_CLASS()
        new.append(
            cls.ENTRY_CLASS(
                RNG.normal(size=100),
                x0=0,
                dx=1,
                dtype=cls.DTYPE,
            ),
        )
        new.append(
            cls.ENTRY_CLASS(
                RNG.normal(size=1000),
                x0=101,
                dx=1,
                dtype=cls.DTYPE,
            ),
        )
        return new

    @pytest.fixture
    def instance(self) -> TimeSeriesBaseListType:
        """Create an instance of a `TimeSeriesList`."""
        return self.create()

    def test_series_link(self):
        """Test that `TimeSeriesList.EntryClass` is set properly."""
        assert self.TEST_CLASS.EntryClass is self.ENTRY_CLASS

    def test_segments(self, instance):
        """Test :attr:`TimeSeriesBaseList.segments`."""
        sl = instance.segments
        assert isinstance(sl, SegmentList)
        assert all(isinstance(s, Segment) for s in sl)
        assert sl == [(0, 100), (101, 1101)]

    def test_append(self, instance):
        """Test `TimeSeriesList.append`."""
        new = self.ENTRY_CLASS([1, 2, 3, 4, 5], x0=1102, dx=1)
        instance.append(new)
        assert len(instance) == 3
        assert instance[-1] is new

    def test_append_typeerror(self, instance):
        """Test `TimeSeriesList.append` errors on type differences."""
        with pytest.raises(
            TypeError,
            match=fr"^Cannot append type 'list' to {self.TEST_CLASS.__name__}$",
        ):
            instance.append([1, 2, 3, 4, 5])

    def test_extend(self):
        """Test `TimeSeriesList.extend`."""
        a = self.create()
        b = a.copy()
        new = self.ENTRY_CLASS([1, 2, 3, 4, 5])
        a.append(new)
        b.extend([new])
        for i in range(max(map(len, (a, b)))):
            utils.assert_quantity_sub_equal(a[i], b[i])

    def test_coalesce(self):
        """Test `TimeSeriesList.coalesce`."""
        a = self.TEST_CLASS()
        a.append(self.ENTRY_CLASS([1, 2, 3, 4, 5], x0=0, dx=1))
        a.append(self.ENTRY_CLASS([1, 2, 3, 4, 5], x0=11, dx=1))
        a.append(self.ENTRY_CLASS([1, 2, 3, 4, 5], x0=5, dx=1))
        a.coalesce()
        assert len(a) == 2
        assert a[0].span == (0, 10)
        utils.assert_array_equal(a[0].value, [1, 2, 3, 4, 5, 1, 2, 3, 4, 5])

    def test_join(self):
        """Test `TimeSeriesList.join`."""
        a = self.TEST_CLASS()
        a.append(self.ENTRY_CLASS([1, 2, 3, 4, 5], x0=0, dx=1))
        a.append(self.ENTRY_CLASS([1, 2, 3, 4, 5], x0=5, dx=1))
        a.append(self.ENTRY_CLASS([1, 2, 3, 4, 5], x0=11, dx=1))

        # disjoint list should throw error
        with pytest.raises(
            ValueError,
            match="Cannot append",
        ):
            a.join()

        # but we can pad to get rid of the errors
        t = a.join(gap="pad")
        assert isinstance(t, a.EntryClass)
        assert t.span == (0, 16)
        utils.assert_array_equal(
            t.value, [1, 2, 3, 4, 5, 1, 2, 3, 4, 5, 0, 1, 2, 3, 4, 5])

    def test_join_empty(self):
        """Test `TimeSeriesList.join` with an empty list."""
        # check that joining empty list produces something sensible
        t = self.TEST_CLASS().join()
        assert isinstance(t, self.TEST_CLASS.EntryClass)
        assert t.size == 0

    def test_slice(self, instance):
        """Test `TimeSeriesList` slicing."""
        s = instance[:2]
        assert type(s) is type(instance)

    def test_copy(self, instance):
        """Test `TimeSeriesList.copy`."""
        a = instance.copy()
        assert type(a) is type(instance)
        for x, y in zip(instance, a, strict=True):
            utils.assert_quantity_sub_equal(x, y)
            assert not shares_memory(x.value, y.value)



# Copyright (c) 2013-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for LAL signal processing interface."""

import pytest
from astropy import units

from ..spectral import _lal as fft_lal

lal = pytest.importorskip("lal")


def test_generate_window():
    """Test :func:`gwpy.signal.spectral.lal.generate_window`."""
    # test default arguments
    w = fft_lal.generate_window(128)
    assert isinstance(w, lal.REAL8Window)
    assert w.data.data.size == 128
    assert w.sum == 32.31817089602309

    # test generating the same window again returns the same object
    assert fft_lal.generate_window(128) is w


def test_generate_window_dtype():
    """Test :func:`gwpy.signal.spectral.lal.generate_window` dtype argument."""
    w = fft_lal.generate_window(128, dtype="float32")
    assert isinstance(w, lal.REAL4Window)
    assert w.sum == 32.31817089602309


@pytest.mark.parametrize(("args", "kwargs", "exc_type"), [
    # invalid window
    ((128,), {"window": "unknown"}, ValueError),
    # unsupported data type
    ((128,), {"dtype": int}, AttributeError),
])
def test_generate_window_error(args, kwargs, exc_type):
    """Test :func:`gwpy.signal.spectral.lal.generate_window` error handling."""
    with pytest.raises(exc_type):
        fft_lal.generate_window(*args, **kwargs)


def test_generate_fft_plan():
    """Test :func:`gwpy.signal.spectral.lal.generate_fft_plan`."""
    # test default arguments
    plan = fft_lal.generate_fft_plan(128)
    assert isinstance(plan, lal.REAL8FFTPlan)
    # test generating the same fft_plan again returns the same object
    assert fft_lal.generate_fft_plan(128) is plan


def test_generate_fft_plan_dtype():
    """Test :func:`gwpy.signal.spectral.lal.generate_fft_plan` dtype."""
    # test dtype works
    plan = fft_lal.generate_fft_plan(128, dtype="float32")
    assert isinstance(plan, lal.REAL4FFTPlan)


def test_generate_fft_plan_forward_backward():
    """Test :func:`gwpy.signal.spectral.lal.generate_fft_plan` forward."""
    # test forward/backward works
    plan = fft_lal.generate_fft_plan(128)
    rvrs = fft_lal.generate_fft_plan(128, forward=False)
    assert isinstance(rvrs, lal.REAL8FFTPlan)
    assert rvrs is not plan


def test_generate_fft_plan_error():
    """Test :func:`gwpy.signal.spectral.lal.generate_fft_plan` error handling."""
    # test errors
    with pytest.raises(AttributeError):
        fft_lal.generate_fft_plan(128, dtype=int)


def test_welch(noisy_sinusoid):
    """Test :func:`gwpy.signal.spectral.lal.welch`."""
    psd = fft_lal.welch(noisy_sinusoid, 4096, noverlap=2048)
    # assert PSD peaks at 500 Hz (as designed)
    assert psd.max() == psd.value_at(500.)
    assert psd.unit == noisy_sinusoid.unit ** 2 / units.Hz
    assert psd.channel is noisy_sinusoid.channel
    assert psd.name is noisy_sinusoid.name


def test_welch_warning(noisy_sinusoid):
    """Test :func:`gwpy.signal.spectral.lal.welch` warnings."""
    # check warning with hanging data
    with pytest.warns(UserWarning):
        fft_lal.welch(noisy_sinusoid, 1000)


def test_median(corrupt_noisy_sinusoid):
    """Test :func:`gwpy.signal.spectral.lal.median`."""
    psd = fft_lal.median(corrupt_noisy_sinusoid, 4096, noverlap=2048)
    assert psd.max() == psd.value_at(500.)
    assert psd.median() < fft_lal.welch(
        corrupt_noisy_sinusoid,
        4096,
        noverlap=2048,
    ).median()


def test_median_mean(corrupt_noisy_sinusoid):
    """Test :func:`gwpy.signal.spectral.lal.median_mean`."""
    psd = fft_lal.median_mean(corrupt_noisy_sinusoid, 8192, noverlap=4096)
    assert psd.max() == psd.value_at(500.)
    assert psd.median() < fft_lal.welch(
        corrupt_noisy_sinusoid,
        4096,
        noverlap=2048,
    ).median()


def test_median_mean_error(corrupt_noisy_sinusoid):
    """Test :func:`gwpy.signal.spectral.lal.median_mean` error handling."""
    # check failure with single segment
    with pytest.raises(ValueError):
        fft_lal.median_mean(
            corrupt_noisy_sinusoid,
            corrupt_noisy_sinusoid.size,
        )



# Copyright (c) 2018-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for :mod:`gwpy.utils.env`."""

from unittest import mock

import pytest

from .. import env as utils_env

BOOL_TRUE = {
    "TEST_y": "y",
    "TEST_Y": "Y",
    "TEST_yes": "yes",
    "TEST_Yes": "Yes",
    "TEST_YES": "YES",
    "TEST_ONE": "1",
    "TEST_true": "true",
    "TEST_True": "True",
    "TEST_TRUE": "TRUE",
}
BOOL_FALSE = {
    "TEST_no": "no",
    "TEST_No": "No",
    "TEST_ZERO": "0",
    "TEST_false": "false",
    "TEST_False": "False",
    "TEST_FALSE": "FALSE",
    "TEST_OTHER": "blah",
}
BOOL_ENV = BOOL_TRUE.copy()
BOOL_ENV.update(BOOL_FALSE)


@mock.patch.dict("os.environ", values=BOOL_ENV)
@pytest.mark.parametrize(
    ("env", "result"),
    [(k, True) for k in sorted(BOOL_TRUE)]
    + [(k, False) for k in sorted(BOOL_FALSE)],
)
def test_bool_env(env, result):
    """Test :meth:`gwpy.utils.env.bool_env` _without_ the `default` keyword."""
    assert utils_env.bool_env(env) is result


@mock.patch.dict("os.environ", values=BOOL_TRUE)
@pytest.mark.parametrize(("env", "default", "result"), [
    pytest.param("TEST_YES", False, True, id="ignore"),
    pytest.param("TEST_MISSING", True, True, id="default"),
])
def test_bool_env_default(env, default, result):
    """Test :meth:`gwpy.utils.env.bool_env` _with_ the `default` keyword."""
    assert utils_env.bool_env(env, default=default) is result



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Command-line utilities for GWpy.

The `gwpy.cli` module provides methods and functionality to power the
`gwpy-plot` command-line executable (distributed with GWpy).
"""

# ruff: noqa: TC001

from __future__ import annotations

from .cliproduct import CliProduct
from .timeseries import TimeSeriesProduct
from .spectrum import SpectrumProduct
from .spectrogram import SpectrogramProduct
from .coherence import CoherenceProduct
from .coherencegram import CoherencegramProduct
from .qtransform import QtransformProduct
from .transferfunction import TransferFunctionProduct

__author__ = "Joseph Areeda <joseph.areeda@ligo.org>"

PRODUCTS: dict[str, type[CliProduct]] = {
    x.action: x
    for x in (
        TimeSeriesProduct,
        SpectrumProduct,
        SpectrogramProduct,
        CoherenceProduct,
        CoherencegramProduct,
        QtransformProduct,
        TransferFunctionProduct,
    )
}



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""
.. sectionauthor:: Duncan Macleod <duncan.macleod@ligo.org>
.. currentmodule:: gwpy.timeseries

Plotting a Rayleigh-statistic `Spectrum`
########################################

In LIGO the 'Rayleigh' statistic is a calculation of the
`coefficient of variation
<https://en.wikipedia.org/wiki/Coefficient_of_variation>`_ of the
power spectral density (PSD) of a given set of data.
It is used to measure the 'Gaussianity' of those data, where a value of 1
indicates Gaussian behaviour, less than 1 indicates coherent variations,
and greater than 1 indicates incoherent variation.
It is a useful measure of the quality of the strain data being generated
and recorded at a LIGO site.
"""

# %%
# To demonstate this, we can load some data from the LIGO Livingston
# interferometer around the time of the GW151226 gravitational wave detection:

from gwpy.timeseries import TimeSeries
gwdata = TimeSeries.get(
    "L1",
    "Dec 26 2015 03:37",
    "Dec 26 2015 03:47",
)

# %%
# Next, we can calculate a Rayleigh statistic `FrequencySeries` using the
# :meth:`~gwpy.timeseries.TimeSeries.rayleigh_spectrum` method of the
# `~gwpy.timeseries.TimeSeries` with a 2-second FFT and 1-second overlap (50%):

rayleigh = gwdata.rayleigh_spectrum(2, 1)

# %%
# For easy comparison, we can calculate the spectral sensitivity ASD of the
# strain data and plot both on the same figure:

from gwpy.plot import Plot
plot = Plot(
    gwdata.asd(2, 1),
    rayleigh,
    geometry=(2, 1),
    sharex=True,
    xscale="log",
    xlim=(30, 1500),
)
asdax, rayax = plot.axes
# update ASD Axes
asdax.set_title("Sensitivity of LIGO-Livingston around GW151226")
asdax.set_yscale("log")
asdax.set_ylim(5e-24, 1e-21)
asdax.set_ylabel(r"[strain/$\sqrt{\mathrm{Hz}}$]")
# update Rayleigh Axes
rayax.set_ylim(0, 2)
rayax.set_ylabel("Rayleigh statistic")
plot.show()

# %%
# So, we see sharp dips at certain frequencies associated with 'lines' in
# spectrum where noise at a fixed frequency is very coherent (e.g. harmonics
# of the 60Hz mains power supply), and bumps in specific frequency bands
# where the interferometer noise is non-stationary.



# Copyright (c) 2024-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Unified I/O read/write for Array objects."""

from __future__ import annotations

from typing import TYPE_CHECKING

from ..io.registry import (
    UnifiedRead,
    UnifiedWrite,
)

if TYPE_CHECKING:
    from collections.abc import Sequence
    from typing import (
        Literal,
        TypeVar,
    )

    from .series import Series

    T = TypeVar("T", bound=Series)


# -- Series --------------------------

class SeriesRead(UnifiedRead):
    """Read data into a `Series`.

    Arguments and keywords depend on the output format, see the
    online documentation for full details for each format, the
    parameters below are common to most formats.

    Parameters
    ----------
    source : `str`, `pathlib.Path`, `file`, `list`
        Source of data, any of the following:

        - open (readable) file object
        - path to a file
        - list of files (open or paths)

    format : `str`,
        Source format identifier. If not given, the format will be
        detected if possible. See below for list of acceptable
        formats.

    kwargs
        Other keywords arguments depend on the format, see the online
        documentation for details.

    Returns
    -------
    data : `Series`

    Raises
    ------
    IndexError
        if ``source`` is an empty list

    Notes
    -----"""

    def merge(  # type: ignore[override]
        self,
        items: Sequence[Series],
        pad: float | None = None,
        gap: Literal["raise", "ignore", "pad"] | None = None,
    ) -> Series:
        """Combine a list of `Series` objects into one `Series`.

        Must be given at least one series.
        """
        itrtr = iter(items)
        out = next(itrtr)
        for series in itrtr:
            out.append(series, gap=gap, pad=pad)
        return out


class SeriesWrite(UnifiedWrite):
    """Write this `Series` to a file.

    Arguments and keywords depend on the output format, see the
    online documentation for full details for each format, the
    parameters below are common to most formats.

    Parameters
    ----------
    target : `str`, `Path`, `file`
        Output file path or file opened in write mode.

    format : `str`, optional
        Output format identifier. Default format will be detected if possible.
        See below for list of acceptable formats.

    Notes
    -----"""


# -- Array2D -------------------------

class Array2DRead(SeriesRead):
    """Read data into a `Array2D`.

    Arguments and keywords depend on the output format, see the
    online documentation for full details for each format, the
    parameters below are common to most formats.

    Parameters
    ----------
    source : `str`, `pathlib.Path`, `file`, `list`
        Source of data, any of the following:

        - open (readable) file object
        - path to a file
        - list of files (open or paths)

    format : `str`,
        Source format identifier. If not given, the format will be
        detected if possible. See below for list of acceptable
        formats.

    kwargs
        Other keywords arguments depend on the format, see the online
        documentation for details.

    Returns
    -------
    data : `Array2D`

    Raises
    ------
    IndexError
        if ``source`` is an empty list

    Notes
    -----"""


class Array2DWrite(UnifiedWrite):
    """Write this `Array2D` to a file.

    Arguments and keywords depend on the output format, see the
    online documentation for full details for each format, the
    parameters below are common to most formats.

    Parameters
    ----------
    target : `str`, `Path`, `file`
        Output file path or file opened in write mode.

    format : `str`, optional
        Output format identifier. Default format will be detected if possible.
        See below for list of acceptable formats.

    Notes
    -----"""



# Copyright (c) 2022 California Institute of Technology
#               2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for :mod:`gwpy.table.io.gstlal`."""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from ...table import EventTable

if TYPE_CHECKING:
    from pathlib import Path

__author__ = "Derek Davis <derek.davis@ligo.org>"


# -- gstlal file fixture -------------

GSTLAL_FILE = """
<?xml version='1.0' encoding='utf-8'?>
<!DOCTYPE LIGO_LW SYSTEM "http://ldas-sw.ligo.caltech.edu/doc/ligolwAPI/html/ligolw_dtd.txt">
<LIGO_LW>
  <Table Name="coinc_inspiral:table">
    <Column Name="coinc_event:coinc_event_id" Type="int_8s"/>
    <Column Name="combined_far" Type="real_8"/>
    <Column Name="end_time" Type="int_4s"/>
    <Column Name="end_time_ns" Type="int_4s"/>
    <Column Name="false_alarm_rate" Type="real_8"/>
    <Column Name="ifos" Type="lstring"/>
    <Column Name="mass" Type="real_8"/>
    <Column Name="mchirp" Type="real_8"/>
    <Column Name="minimum_duration" Type="real_8"/>
    <Column Name="snr" Type="real_8"/>
    <Stream Name="coinc_inspiral:table" Delimiter="," Type="Local">
      1,1,100,0,1,"H1,L1",1,1,1,1,
      2,1,100,0,1,"H1,L1",1,1,1,1,
    </Stream>
  </Table>
  <Table Name="coinc_event:table">
    <Column Name="coinc_definer:coinc_def_id" Type="int_8s"/>
    <Column Name="coinc_event_id" Type="int_8s"/>
    <Column Name="instruments" Type="lstring"/>
    <Column Name="likelihood" Type="real_8"/>
    <Column Name="nevents" Type="int_4u"/>
    <Column Name="process:process_id" Type="int_8s"/>
    <Column Name="time_slide:time_slide_id" Type="int_8s"/>
    <Stream Name="coinc_event:table" Delimiter="," Type="Local">
      0,1,"H1,L1",1,1,0,0,
      1,1,"H1,L1",1,1,0,0
    </Stream>
  </Table>
  <Table Name="sngl_inspiral:table">
    <Column Name="process:process_id" Type="int_8s"/>
    <Column Name="ifo" Type="lstring"/>
    <Column Name="end_time" Type="int_4s"/>
    <Column Name="end_time_ns" Type="int_4s"/>
    <Column Name="eff_distance" Type="real_4"/>
    <Column Name="coa_phase" Type="real_4"/>
    <Column Name="mass1" Type="real_4"/>
    <Column Name="mass2" Type="real_4"/>
    <Column Name="snr" Type="real_4"/>
    <Column Name="chisq" Type="real_4"/>
    <Column Name="chisq_dof" Type="int_4s"/>
    <Column Name="bank_chisq" Type="real_4"/>
    <Column Name="bank_chisq_dof" Type="int_4s"/>
    <Column Name="sigmasq" Type="real_8"/>
    <Column Name="spin1x" Type="real_4"/>
    <Column Name="spin1y" Type="real_4"/>
    <Column Name="spin1z" Type="real_4"/>
    <Column Name="spin2x" Type="real_4"/>
    <Column Name="spin2y" Type="real_4"/>
    <Column Name="spin2z" Type="real_4"/>
    <Column Name="template_duration" Type="real_8"/>
    <Column Name="event_id" Type="int_8s"/>
    <Column Name="Gamma0" Type="real_4"/>
    <Column Name="Gamma1" Type="real_4"/>
    <Column Name="Gamma2" Type="real_4"/>
    <Stream Name="sngl_inspiral:table" Delimiter="," Type="Local">
      1,"L1",100,0,nan,0,10,10,3,2,1,1,1,1,0,0,0,0,0,0,1,1,1,1,1,
      1,"H1",100,0,nan,0,10,10,3,2,1,1,1,1,0,0,0,0,0,0,1,1,1,1,1,
      1,"V1",100,0,nan,0,10,10,3,2,1,1,1,1,0,0,0,0,0,0,1,1,1,1,1
    </Stream>
  </Table>
</LIGO_LW>
"""


@pytest.fixture
def gstlal_table(tmp_path: Path) -> Path:
    """Create a temporary gstlal XML file."""
    tmp = tmp_path / "H1L1V1-LLOID-1-1.xml.gz"
    tmp.write_text(GSTLAL_FILE.strip())
    return tmp


# -- tests ---------------------------

SNGL_LEN = 3
COINC_LEN = 2


@pytest.mark.requires("igwn_ligolw.lsctables")
def test_read_gstlal_sngl(gstlal_table: Path):
    """Test reading ``sngl_inspiral`` triggers from a gstlal file."""
    table = EventTable.read(
        gstlal_table,
        format="ligolw.gstlal",
        triggers="sngl",
    )
    assert len(table) == SNGL_LEN


@pytest.mark.requires("igwn_ligolw.lsctables")
def test_read_sngl_format(gstlal_table: Path):
    """Test reading ``sngl_inspiral`` triggers  with ``format``."""
    table = EventTable.read(gstlal_table, format="ligolw.gstlal.sngl")
    assert len(table) == SNGL_LEN


@pytest.mark.requires("igwn_ligolw.lsctables")
def test_read_sngl_columns(gstlal_table):
    """Test reading ``sngl_inspiral`` triggers with specified columns."""
    table = EventTable.read(
        gstlal_table,
        format="ligolw.gstlal.sngl",
        columns=["snr", "end_time"],
    )
    assert set(table.keys()) == {"snr", "end_time"}


@pytest.mark.requires("igwn_ligolw.lsctables")
def test_read_coinc(gstlal_table):
    """Test reading ``coinc_inspiral`` triggers from a gstlal file."""
    table = EventTable.read(
        gstlal_table,
        format="ligolw.gstlal",
        triggers="coinc",
    )
    assert len(table) == COINC_LEN


@pytest.mark.requires("igwn_ligolw.lsctables")
def test_read_coinc_format(gstlal_table):
    """Test reading ``coinc_inspiral`` triggers with ``format``."""
    table = EventTable.read(
        gstlal_table,
        format="ligolw.gstlal.coinc",
    )
    assert len(table) == COINC_LEN


@pytest.mark.requires("igwn_ligolw.lsctables")
def test_read_coinc_columns(gstlal_table: Path):
    """Test reading ``coinc_inspiral`` triggers with specified columns."""
    table = EventTable.read(
        gstlal_table,
        format="ligolw.gstlal.coinc",
        columns=["snr", "end_time"],
    )
    assert set(table.keys()) == {"snr", "end_time"}


@pytest.mark.requires("igwn_ligolw.lsctables")
def test_derived_values(gstlal_table: Path):
    """Test reading ``sngl_inspiral`` triggers with derived values."""
    table = EventTable.read(
        gstlal_table,
        format="ligolw.gstlal",
        triggers="sngl",
        columns=["snr_chi", "chi_snr"],
    )
    assert table["snr_chi"][0] == pytest.approx(4.5)
    assert table["chi_snr"][0] == pytest.approx(1./4.5)
    assert "snr" not in table.colnames


@pytest.mark.requires("igwn_ligolw.lsctables")
def test_incorrect_sngl_column(gstlal_table: Path):
    """Test reading ``sngl_inspiral`` triggers with incorrect column names."""
    with pytest.raises(
        ValueError,
        match="is not a valid column name",
    ):
        EventTable.read(
            gstlal_table,
            format="ligolw.gstlal.sngl",
            columns=["nan"],
        )


@pytest.mark.requires("igwn_ligolw.lsctables")
def test_incorrect_coinc_column(gstlal_table: Path):
    """Test reading ``coinc_inspiral`` triggers with incorrect column names."""
    with pytest.raises(
        ValueError,
        match="is not a valid column name",
    ):
        EventTable.read(
            gstlal_table,
            format="ligolw.gstlal.coinc",
            columns=["nan"],
       )


@pytest.mark.requires("igwn_ligolw.lsctables")
def test_incorrect_trigger_name(gstlal_table: Path):
    """Test reading gstlal file with incorrect trigger type."""
    with pytest.raises(
        ValueError,
        match=r"^triggers must be 'sngl' or 'coinc', got 'nan'$",
    ):
        EventTable.read(
            gstlal_table,
            format="ligolw.gstlal",
            triggers="nan",
        )



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Read/write WAV files using `scipy.signal.wavfile`."""

from __future__ import annotations

import struct
import wave
from typing import TYPE_CHECKING

import numpy
from scipy.io import wavfile

from .. import TimeSeries

if TYPE_CHECKING:
    from pathlib import Path
    from typing import IO

    from ...time import SupportsToGps

WAV_SIGNATURE = ("RIFF", "WAVE")


def read(
    fobj: str | Path | IO,
    start: SupportsToGps | None = None,
    end: SupportsToGps | None = None,
    **kwargs,
) -> TimeSeries:
    """Read a WAV file into a `TimeSeries`.

    Parameters
    ----------
    fobj : `file`, `str`
        Open file-like object or filename to read from.

    start : `float`, `astropy.units.Quantity`, optional
        The desired start point of the X-axis, defaults to
        the start point of the incoming series.

    end : `float`, `astropy.units.Quantity`, optional
        The desired end point of the X-axis, defaults to
        the end point of the incoming series.

    **kwargs
        All keyword arguments are passed onto :func:`scipy.io.wavfile.read`.

    See Also
    --------
    scipy.io.wavfile.read
        For details on how the WAV file is actually read.

    Examples
    --------
    >>> from gwpy.timeseries import TimeSeries
    >>> t = TimeSeries.read('test.wav')
    """
    fsamp, arr = wavfile.read(fobj, **kwargs)
    return TimeSeries(arr, sample_rate=fsamp).crop(
        start=start,
        end=end,
    )


def write(
    series: TimeSeries,
    output: str | IO,
    scale: float | None = None,
) -> None:
    """Write a `TimeSeries` to a WAV file.

    Parameters
    ----------
    series : `TimeSeries`
        the series to write

    output : `file`, `str`
        the file object or filename to write to

    scale : `float`, optional
        the factor to apply to scale the data to (-1.0, 1.0),
        pass `scale=1` to not apply any scale, otherwise
        the data will be auto-scaled

    See Also
    --------
    scipy.io.wavfile.write
        for details on how the WAV file is actually written

    Examples
    --------
    >>> from gwpy.timeseries import TimeSeries
    >>> t = TimeSeries([1, 2, 3, 4, 5])
    >>> t = TimeSeries.write('test.wav')
    """
    fsamp = int(series.sample_rate.decompose().value)
    if scale is None:
        scale = 1 / numpy.abs(series.value).max()
    data = (series.value * scale).astype("float32")
    return wavfile.write(output, fsamp, data)


# -- identify ------------------------

def _is_wav(
    origin: str,
    filepath: str,
    fileobj: IO,
    *args: IO,
    **kwargs,
) -> bool:
    """Identify a file as WAV.

    See `astropy.io.registry` for details on how this function is used.
    """
    # if given an open, readable file object, inspect the magic signature
    if origin == "read" and fileobj is not None:
        loc = fileobj.tell()
        fileobj.seek(0)
        try:
            riff, _, fmt = struct.unpack("<4sI4s", fileobj.read(12))
            if isinstance(riff, bytes):
                riff = riff.decode("utf-8")
                fmt = fmt.decode("utf-8")
            return riff == WAV_SIGNATURE[0] and fmt == WAV_SIGNATURE[1]
        except (
            UnicodeDecodeError,
            struct.error,
        ):
            return False
        finally:
            fileobj.seek(loc)

    # check the file extension
    if filepath is not None:
        return filepath.endswith((".wav", ".wave"))

    # attempt to read the file as .wav
    kwargs.pop("start", None)
    kwargs.pop("end", None)
    try:
        with wave.open(*args, **kwargs):
            return True
    except (
        AttributeError,  # not a file-like object
        TypeError,  # wave.open failed for some other reason
        wave.Error,  # scipy failed to read this as a WAV file
    ):
        return False


# -- register ------------------------

registry = TimeSeries.read.registry
registry.register_reader("wav", TimeSeries, read)
registry.register_writer("wav", TimeSeries, write)
registry.register_identifier("wav", TimeSeries, _is_wav)



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Common utilities for frequency-domain operations.

This module holds code used by both the `FrequencySeries` and `Spectrogram`.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

import numpy

from ..signal import filter_design

if TYPE_CHECKING:
    from typing import TypeVar

    from astropy.units.typing import QuantityLike

    from ..frequencyseries import FrequencySeries
    from ..signal.filter_design import FilterCompatible
    from ..spectrogram import Spectrogram

    FreqDomainData = TypeVar("FreqDomainData", FrequencySeries, Spectrogram)

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"


def _fdfilter(
    data: FreqDomainData,
    filt: FilterCompatible,
    *,
    sample_rate: QuantityLike | None = None,
    analog: bool = False,
    unit: str = "rad/s",
    normalize_gain: bool = False,
    inplace: bool = False,
    **kwargs,
) -> FreqDomainData:
    """Filter a frequency-domain data object.

    See Also
    --------
    gwpy.frequencyseries.FrequencySeries.filter
    gwpy.spectrogram.Spectrogram.filter
    """
    freqs = data.frequencies.to("Hz").value
    if sample_rate is None:
        sample_rate = freqs[-1] * 2

    _, fresp = numpy.abs(filter_design.frequency_response(
        filt,
        freqs,
        analog=analog,
        sample_rate=sample_rate,
        unit=unit,
        normalize_gain=normalize_gain,
        **kwargs,
    ))

    # apply to array
    if inplace:
        data *= fresp
        return data
    return data * fresp



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for :mod:`gwpy.segments.segments`."""

from __future__ import annotations

from typing import TYPE_CHECKING

import h5py
import pytest
from astropy.table import Table

from ...testing.errors import pytest_skip_flaky_network
from ...testing.utils import (
    TEST_DATA_PATH,
    assert_segmentlist_equal,
    assert_table_equal,
)
from ...time import LIGOTimeGPS
from .. import (
    Segment,
    SegmentList,
)

if TYPE_CHECKING:
    from typing import SupportsFloat

TEST_SEGWIZARD_FILE = TEST_DATA_PATH / "X1-GWPY_TEST_SEGMENTS-0-10.txt"
TEST_SEGWIZARD_URI = (
    "https://gitlab.com/gwpy/gwpy/-/raw/v3.0.10/"
    + TEST_SEGWIZARD_FILE.relative_to(
        TEST_DATA_PATH.parent.parent.parent,
    ).as_posix()
)


def _as_segmentlist(*segments: tuple[SupportsFloat, SupportsFloat]) -> SegmentList:
    """Return ``segments`` as a `SegmentList`."""
    return SegmentList([Segment(a, b) for a, b in segments])


# -- Segment -------------------------

class TestSegment:
    """Test `gwpy.segments.Segment`."""

    TEST_CLASS = Segment

    @pytest.fixture
    @classmethod
    def segment(cls):
        """Create a test segment (fixture)."""
        return cls.TEST_CLASS(1, 2)

    def test_start_end(self, segment):
        """Test the start and end properties."""
        assert segment.start == 1.
        assert segment.end == 2.

    def test_repr(self, segment):
        """Test ``repr(segment)``."""
        assert repr(segment) == "Segment(1, 2)"

    def test_str(self, segment):
        """Test ``str(segment)``."""
        assert str(segment) == "[1 ... 2)"


# -- SegmentList ---------------------

class TestSegmentList:
    """Test `gwpy.segments.SegmentList`."""

    TEST_CLASS = SegmentList
    ENTRY_CLASS = Segment

    @classmethod
    def create(cls, *segments):
        """Create a test segment list."""
        return cls.TEST_CLASS([cls.ENTRY_CLASS(a, b) for a, b in segments])

    @pytest.fixture
    @classmethod
    def segmentlist(cls):
        """Create a test segment list (fixture)."""
        return cls.create((1, 2), (3, 4), (4, 6), (8, 10))

    # -- test methods ----------------

    def test_extent(self, segmentlist):
        """Test `SegmentList.extent()` returns the right type."""
        extent = segmentlist.extent()
        assert isinstance(extent, self.ENTRY_CLASS)
        assert extent == Segment(1, 10)

    def test_coalesce(self):
        """Test `SegmentList.coalesce()`."""
        segmentlist = self.create((1, 2), (3, 4), (4, 5))
        c = segmentlist.coalesce()
        assert c is segmentlist
        assert_segmentlist_equal(c, _as_segmentlist((1, 2), (3, 5)))
        assert isinstance(c[0], self.ENTRY_CLASS)

    def test_to_table(self, segmentlist):
        """Test `SegmentList.to_table()`."""
        segtable = segmentlist.to_table()
        assert_table_equal(
            segtable,
            Table(
                rows=[
                    (0, 1, 2, 1),
                    (1, 3, 4, 1),
                    (2, 4, 6, 2),
                    (3, 8, 10, 2),
                ],
                names=("index", "start", "end", "duration"),
            ),
        )

    # -- test I/O --------------------

    def test_read_write_segwizard(self, segmentlist, tmp_path):
        """Test writing and reading back a `SegmentList` in ASCII format."""
        tmp = tmp_path / "segments.txt"
        # check write/read returns the same list
        segmentlist.write(tmp)
        sl2 = self.TEST_CLASS.read(tmp, coalesce=False)
        assert_segmentlist_equal(sl2, segmentlist)
        assert isinstance(sl2[0][0], LIGOTimeGPS)

        # check that coalesceing does what its supposed to
        c = type(segmentlist)(segmentlist).coalesce()
        sl2 = self.TEST_CLASS.read(str(tmp), coalesce=True)
        assert_segmentlist_equal(sl2, c)

        # check gpstype kwarg
        sl2 = self.TEST_CLASS.read(tmp, gpstype=float)
        assert isinstance(sl2[0][0], float)

    def test_read_write_segwizard_strict(self, tmp_path):
        """Test writing and reading back a `SegmentList` in strict ASCII format."""
        tmp = tmp_path / "segments.txt"
        tmp.write_text("0 0 1 .5")
        with pytest.raises(
            ValueError,
            match=r"incorrect duration \.5",
        ):
            self.TEST_CLASS.read(tmp, strict=True, format="segwizard")
        sl = self.TEST_CLASS.read(tmp, strict=False, format="segwizard")
        assert_segmentlist_equal(sl, _as_segmentlist((0, 1)))

    def test_read_write_segwizard_twocol(self, tmp_path):
        """Test writing and reading back a `SegmentList` in two-col ASCII format."""
        tmp = tmp_path / "segments.txt"
        tmp.write_text("0 1\n2 3")
        sl = self.TEST_CLASS.read(tmp, format="segwizard")
        assert_segmentlist_equal(sl, _as_segmentlist((0, 1), (2, 3)))

    @pytest.mark.parametrize("ext", [".hdf5", ".h5"])
    def test_read_write_hdf5(self, segmentlist, tmp_path, ext):
        """Test writing and reading back a `SegmentList` in HDF5 format."""
        tmp = tmp_path / f"segments{ext}"

        # check basic write/read with auto-path discovery
        segmentlist.write(tmp, "test-segmentlist")
        sl2 = self.TEST_CLASS.read(tmp)
        assert_segmentlist_equal(sl2, segmentlist)
        assert isinstance(sl2[0][0], LIGOTimeGPS)

        sl2 = self.TEST_CLASS.read(tmp, path="test-segmentlist")
        assert_segmentlist_equal(sl2, segmentlist)

        # check we can read directly from the h5 object
        with h5py.File(tmp, "r") as h5f:
            sl2 = self.TEST_CLASS.read(h5f["test-segmentlist"])
            assert_segmentlist_equal(sl2, segmentlist)

        # check overwrite kwarg
        with pytest.raises(
            OSError,
            match=r"File .* already exists",
        ):
            segmentlist.write(tmp, "test-segmentlist")
        segmentlist.write(tmp, "test-segmentlist", overwrite=True)

        # check gpstype kwarg
        sl2 = self.TEST_CLASS.read(tmp, gpstype=float)
        assert_segmentlist_equal(sl2, segmentlist)
        assert isinstance(sl2[0][0], float)

    @pytest_skip_flaky_network
    def test_read_remote(self):
        """Test that reading directly from a remote URI works."""
        local = self.TEST_CLASS.read(TEST_SEGWIZARD_FILE)
        remote = self.TEST_CLASS.read(TEST_SEGWIZARD_URI, cache=False)
        assert_segmentlist_equal(local, remote)



# Copyright (c) 2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Registry utilities for Table I/O."""

from __future__ import annotations

from functools import wraps
from typing import TYPE_CHECKING

from .. import EventTable
from .utils import (
    read_with_columns,
    read_with_where,
)

if TYPE_CHECKING:
    from astropy.io.registry import UnifiedIORegistry
    from astropy.table import Table

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"


def decorate_registered_reader(
    name: str,
    data_class: type[Table] = EventTable,
    *,
    columns: bool = True,
    where: bool = True,
    registry: UnifiedIORegistry | None = None,
) -> None:
    """Wrap an existing registered reader to use GWpy's input decorators.

    Parameters
    ----------
    name : `str`
        The name of the registered format.

    data_class : `type`
        The class for whom the format is registered.

    columns : `bool`
        Use the `read_with_columns` decorator.

    where : `bool`
        Use the `read_with_where` decorator.

    registry : `astropy.io.registry.UnifiedIORegistry`, optional
        The registry to modify.
        If `None` (default) then use the registry for `data_class`.
    """
    if registry is None:
        registry = data_class.read.registry
    orig = reader = registry.get_reader(name, data_class)
    if where:
        reader = wraps(orig)(read_with_where(reader))
    if columns:
        reader = wraps(orig)(read_with_columns(reader))
    return registry.register_reader(
        name,
        data_class,
        reader,
        force=True,
    )


def wrap_unified_io_readers(klass: type[Table]) -> None:
    """Decorate registered readers of ``klass`` to support ``columns`` and ``where``.

    See Also
    --------
    read_with_columns
        For details of the decorator adding ``columns`` keyword support.

    read_with_where
        For details of the decorator adding ``where`` keyword support.
    """
    registry = klass.read.registry
    # wrap each reader to support columns and where keywords
    for row in registry.get_formats(
        data_class=klass,
        readwrite="Read",
    ):
        decorate_registered_reader(
            row["Format"],
            data_class=klass,
            columns=True,
            where=True,
            registry=registry,
        )



# Copyright (c) 2024-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Remote file I/O utilities for GWpy.

This module provides the `open_remote_file` function, which is applied
automatically when reading all files with the `gwpy.io.registry.UnifiedRead`
registry reader.
This means that operations like `TimeSeries.read("http://...")` can access
remote data.

The `open_remote_file` function also handles (or tries to handle) Pelican
URLs, via a hand-off to the function of the same name in `gwpy.io.pelican`.
"""

from __future__ import annotations

import logging
import socket
from pathlib import PureWindowsPath
from ssl import SSLError
from typing import (
    TYPE_CHECKING,
    overload,
)
from unittest import mock
from urllib.error import URLError

import requests.exceptions
from astropy.utils import data as astropy_data
from urllib3.util import parse_url

from ..utils.env import bool_env
from . import pelican as io_pelican

if TYPE_CHECKING:
    from contextlib import AbstractContextManager
    from typing import (
        IO,
        BinaryIO,
        Literal,
        TextIO,
    )

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"
__all__ = [
    "download_file",
    "is_remote",
    "open_remote_file",
]

log = logging.getLogger(__name__)

#: Tuple of exception types that indicate a network error.
#:
#: Used to catch network errors when attempting to download remote files.
NETWORK_ERROR: tuple[type[Exception], ...] = (
    ConnectionError,
    requests.exceptions.ConnectionError,
    requests.exceptions.ReadTimeout,
    socket.timeout,
    SSLError,
    TimeoutError,
    URLError,
)

# -- Patch astropy to log downloads

_download_file_from_source = astropy_data._download_file_from_source

def _logging_download_file_from_source(url: str, *args, **kwargs) -> str:
    """Patch for astropy's _download_file_from_source to log the URL.

    This is helpful for debugging downloads with multiple sources or
    redirects.
    """
    log.debug("Downloading file from %s", url)
    out = _download_file_from_source(url, *args, **kwargs)
    log.debug("Downloaded file from %s to %s", url, out)
    return out


def _mock_download_file_from_source() -> AbstractContextManager:
    """Context manager to mock astropy's internal download function to log URLs."""
    return mock.patch(
        "astropy.utils.data._download_file_from_source",
        _logging_download_file_from_source,
    )


# -- Remote file handling

def is_remote(url: str) -> bool:
    r"""Return `True` if ``url`` points at a remote URL.

    This function just inspects the 'scheme' of the URL, if present,
    and returns `True` if the scheme *isn't* ``"file"``.

    On Windows, any ``url`` that includes a drive assignment, e.g:
    ``"C:\\Users\\me\\data.txt"`` will return `False`.

    Parameters
    ----------
    url : `str`, `pathlib.Path`
        The URL to inspect.

    Returns
    -------
    remote : `bool`
        `True` if the URL looks like it is remote (would need a
        network connection for access), otherwise `False`.
    """
    if PureWindowsPath(url).drive:  # windows drive path
        return False
    # check the URL scheme
    return parse_url(url).scheme not in (None, "file")


@overload
def open_remote_file(
    url: str,
    *,
    cache: bool | None = None,
    show_progress: bool = False,
    encoding: Literal["binary"] = "binary",
    **kwargs,
) -> AbstractContextManager[BinaryIO]: ...

@overload
def open_remote_file(
    url: str,
    *,
    cache: bool | None = None,
    show_progress: bool = False,
    encoding: str | None,
    **kwargs,
) -> AbstractContextManager[TextIO]: ...

def open_remote_file(
    url: str,
    *,
    cache: bool | None = None,
    show_progress: bool = False,
    encoding: str | None = "binary",
    **kwargs,
) -> AbstractContextManager[IO]:
    """Download a file and open it.

    This function is a wrapper around `astropy.utils.data.get_readable_fileobj`
    with the following customisations:

    - Default to ``cache=True`` if the ``GWPY_CACHE`` environment variable
      is set to something 'truthy'.

    - If the URL looks like a |Pelican|_ URL, hand off to a dedicated
      Pelican download wrapper to attempt the remote access.

    Parameters
    ----------
    url : `str`, file-like
        The name of the resource to access. Can be a local path, or a
        `file://` URL, or any remote URL supported by Astropy.

    cache : `bool`
        Whether to cache the contents of remote URLs.
        Default is `True` if the ``GWPY_CACHE`` environment variable
        is set to something 'truthy'.

    show_progress: `bool`, optional
        Print verbose progress information to the screen.

    encoding : `str`, optional
        When ``'binary'`` (default), returns a file-like object where its
        ``read`` method returns `bytes` objects.

        When `None`, returns a file-like object with a
        ``read`` method that returns `str` (``unicode``) objects, using
        `locale.getpreferredencoding` as an encoding.  This matches
        the default behavior of the built-in `open` when no ``mode``
        argument is provided.

        When another string, it is the name of an encoding, and the
        file-like object's ``read`` method will return `str` (``unicode``)
        objects, decoded from binary using the given encoding.

    kwargs
        All other positional and keyword arguments are passed directly
        to `astropy.utils.data.get_readable_fileobj`.

    Returns
    -------
    file : file-like
        The file opened in binary format.

    See Also
    --------
    astropy.utils.data.get_readable_fileobj
        For details of the underlying downloader.

    gwpy.io.pelican.open_remote_file
        For details of the Pelican-aware download wrapper.
    """
    return _handle_remote_file(
        url,
        mode="open",
        cache=cache,
        show_progress=show_progress,
        encoding=encoding,
        **kwargs,
    )


def download_file(
    url: str,
    *,
    cache: bool | None = None,
    show_progress: bool = False,
    **kwargs,
) -> str:
    """Download a file from a URL and optionally cache the result.

    This function is a wrapper around `astropy.utils.data.download_file`
    with the following customisations:

    - File URLs (``file://``) and plain paths (without a URL scheme)
      are returned unmodified **without** being downloaded or cached.

    - Default to ``cache=True`` if the ``GWPY_CACHE`` environment variable
      is set to something 'truthy'.

    - If the URL looks like a |Pelican|_ URL, hand off to a dedicated
      Pelican download wrapper to attempt the remote access.

    Parameters
    ----------
    url : `str`, file-like
        The name of the resource to access. Can be a local path, or a
        `file://` URL, or any remote URL supported by Astropy.

    cache : `bool`
        Whether to cache the contents of remote URLs.
        Default is `True` if the ``GWPY_CACHE`` environment variable
        is set to something 'truthy'.

    show_progress: `bool`, optional
        Print verbose progress information to the screen.

    kwargs
        All other positional and keyword arguments are passed directly
        to `astropy.utils.data.get_readable_fileobj`.

    Returns
    -------
    load_path : str
        The local path that the file was downloaded to.

    See Also
    --------
    astropy.utils.data.download_file
        For details of the underlying downloader.

    gwpy.io.pelican.open_remote_file
        For details of the Pelican-aware download wrapper.
    """
    return _handle_remote_file(
        url,
        mode="download",
        cache=cache,
        show_progress=show_progress,
        **kwargs,
    )


@overload
def _handle_remote_file(
    url: str,
    mode: Literal["open"],
    *,
    cache: bool | None = None,
    **kwargs,
) -> AbstractContextManager[BinaryIO]: ...

@overload
def _handle_remote_file(
    url: str,
    mode: Literal["download"],
    *,
    cache: bool | None = None,
    **kwargs,
) -> str: ...

def _handle_remote_file(
    url: str,
    mode: Literal["open", "download"] = "open",
    *,
    cache: bool | None = None,
    **kwargs,
) -> AbstractContextManager[IO] | str:
    """Handle getting a remote file, including Pelican URLs.

    This just abstracts out the common code used by both
    `open_remote_file` and `download_file`.
    """
    # enable caching by default based on environment
    if mode == "download" and not is_remote(url):
        return url

    if mode == "download":
        get = astropy_data.download_file
        pelican_get = io_pelican.download_file
    else:
        get = astropy_data.get_readable_fileobj
        pelican_get = io_pelican.open_remote_file  # type: ignore[assignment]

    if cache is None:
        cache = bool_env("GWPY_CACHE", default=False)

    with _mock_download_file_from_source():
        # if given a Pelican URL hand off to the Pelican-aware wrapper
        if io_pelican.is_pelican_url(url):
            return pelican_get(
                url,
                cache=cache,
                **kwargs,
            )

        # download the file
        # note: get_readable_fileobj is a context manager, so doesn't execute here
        return get(
            url,
            cache=cache,
            **kwargs,
        )



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2021 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""I/O registry extensions on top of `astropy.io.registry`.

This module imports a subset of the useful functions from
:mod:`astropy.io.registry` for convenience.
"""

from __future__ import annotations

import contextlib
import inspect
import logging
import os
import pydoc
import re
import warnings
from abc import (
    ABC,
    abstractmethod,
)
from concurrent.futures import (
    ThreadPoolExecutor,
    as_completed,
)
from contextlib import nullcontext
from typing import (
    TYPE_CHECKING,
    Generic,
    Protocol,
    TypeVar,
    cast,
)

from astropy.io import registry as astropy_registry
from astropy.io.registry import (
    UnifiedReadWriteMethod,  # noqa: F401
)

from ..utils.env import bool_env
from ..utils.progress import progress_bar
from .remote import open_remote_file
from .utils import (
    file_list,
)

if TYPE_CHECKING:
    from collections.abc import (
        Callable,
        Iterable,
        Sequence,
    )
    from contextlib import AbstractContextManager
    from typing import (
        Any,
        BinaryIO,
        Literal,
        Never,
    )

    from .utils import (
        FileLike,
        FileSystemPath,
        NamedReadable,
    )

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"
__all__ = [
    "GetExceptionGroup",
    "UnifiedFetch",
    "UnifiedFetchRegistry",
    "UnifiedGet",
    "UnifiedGetRegistry",
    "UnifiedIORegistry",
    "UnifiedRead",
    "UnifiedWrite",
    "default_registry",
    "identify_factory",
    "inherit_unified_io",
]

# Type variable for Generic classes
T = TypeVar("T")


class GetExceptionGroup(ExceptionGroup):
    """Exception group raised by ``Klass.get()`` when all sources fail.

    This is a subclass of `ExceptionGroup` that can be used to identify
    exceptions raised specifically by the ``Klass.get()`` class when all
    sources fail to get data.
    """


# -- Identify utilities --------------

class IdentifyProtocol(Protocol):
    """Typing protocol for Astropy I/O identify functions."""

    def __call__(
        self,
        origin: Literal["read", "write"],
        filepath: FileSystemPath | None,
        fileobj: FileLike | None,
        *args,  # noqa: ANN002
        **kwargs,
    ) -> bool:
        """Identify the given extensions in a file object/path."""


def identify_factory(*extensions: str) -> IdentifyProtocol:
    """Return an Astropy Unified I/O identify function for a set of extensions.

    The returned function is designed for use in the unified I/O registry
    via the `astropy.io.registry.register_identifier` hook.

    Parameters
    ----------
    extensions : `str`
        one or more file extension strings

    Returns
    -------
    identifier : `callable`
        an identifier function that tests whether an incoming file path
        carries any of the given file extensions (using `str.endswith`)
    """
    def identify(
        origin: Literal["read", "write"],  # noqa: ARG001
        filepath: FileSystemPath | None,
        fileobj: FileLike | None,  # noqa: ARG001
        *args,  # noqa: ANN002, ARG001
        **kwargs,  # noqa: ARG001
    ) -> bool:
        """Identify the given extensions in a file object/path."""
        return (
            filepath is not None
            and str(filepath).endswith(extensions)
        )
    return identify


def _list_identifier(
    identifier: IdentifyProtocol,
) -> IdentifyProtocol:
    """Decorate an I/O identifier to handle a list of files as input.

    This function tries to resolve a single file path as a `str` from any
    one or more file-like objects in the ``filepath`` or ``args`` inputs
    to pass to the underlying identifier for comparison.
    """
    def decorated_func(
        origin: Literal["read", "write"],
        filepath: FileSystemPath | None,
        fileobj: FileLike | None,
        *args,  # noqa: ANN002
        **kwargs,
    ) -> bool:
        target = filepath  # thing to search
        if target is None and args:
            target = args[0]
        with contextlib.suppress(
            IndexError,  # empty list
            ValueError,  # target can't be resolved as a list of file-like
        ):
            filepath = file_list(target)[0]  # type: ignore[arg-type]
        return identifier(origin, filepath, fileobj, *args, **kwargs)

    return decorated_func


# -- Unified I/O format --------------

class UnifiedIORegistry(astropy_registry.UnifiedIORegistry):
    """UnifiedIORegistry that can handle reading files in parallel."""

    def identify_format(
        self,
        origin: Literal["read", "write"],
        data_class_required: type,
        path: FileSystemPath | None,
        fileobj: FileLike | None,
        args: tuple,
        kwargs: dict,
    ) -> list[str]:
        """Identify the format of a file, handling lists of files."""
        if fileobj is None:
            with contextlib.suppress(
                IndexError,  # list is empty
                ValueError,  # failed to parse as list-like
            ):
                path = file_list(path)[0]  # type: ignore[arg-type]
        return super().identify_format(
            origin,
            data_class_required,
            path,
            fileobj,
            args,
            kwargs,
        )

    identify_format.__doc__ = (
        astropy_registry.UnifiedIORegistry.identify_format.__doc__
    )

    def register_identifier(
        self,
        data_format: str,
        data_class: type,
        identifier: IdentifyProtocol,
        force: bool = False,  # noqa: FBT001,FBT002
    ) -> None:
        """Register an identifier function that can handle lists of files."""
        return super().register_identifier(
            data_format,
            data_class,
            _list_identifier(identifier),
            force=force,
        )

    register_identifier.__doc__ = (
        astropy_registry.UnifiedIORegistry.register_identifier.__doc__
    )


#: The default unified I/O registry instance.
#:
#: Most classes should use this registry unless they have a specific reason
#: to use a custom registry.
default_registry = UnifiedIORegistry()



class MergeProtocol(Protocol[T]):
    """Typing protocol for merge functions used in `UnifiedRead`."""

    def __call__(self, items: Sequence[T]) -> T:
        """Merge a sequence of items into a single item."""


class UnifiedRead(astropy_registry.UnifiedReadWrite, ABC, Generic[T]):
    """Base ``Class.read()`` implementation that handles parallel reads.

    Each parallel read must return an instance of the same type
    as the input class, and the results are merged by the provided
    ``merge_function`` which should have the following signature::

        def merge_function(self: Type, instances: list[Type], **kwargs) -> Type

    i.e take in the type object and a list of instances, and return
    a single instance of the same type.
    """

    def __init__(
        self,
        instance: object,
        cls: type[T],
        registry: UnifiedIORegistry = default_registry,
    ) -> None:
        """Initialise a new `UnifiedRead` instance."""
        super().__init__(
            instance,
            cls,
            "read",
            registry=registry,
        )

    # -- Merge -----------------------

    @abstractmethod
    def merge(self, items: Sequence[T], **kwargs) -> T:
        """Merge a sequence of items into a single item.

        Parameters
        ----------
        items : Sequence[T]
            The sequence of items to merge.
        **kwargs
            Additional keyword arguments specific to the merge operation.
            Subclasses may override this method with specific named parameters;
            use ``# type: ignore[override]`` to suppress mypy warnings about
            signature compatibility.

        Returns
        -------
        T
            The merged item.

        Notes
        -----
        The base class automatically extracts merge-specific kwargs from the
        ``__call__`` method by inspecting the merge method signature, so
        subclasses with specific named parameters will receive them correctly.
        """

    def _merge_kwargs(self, kwargs: dict, merge_function: MergeProtocol) -> dict:
        merge_sig = inspect.signature(merge_function)
        return {
            key: kwargs.pop(key)
            for key in list(kwargs.keys())
            if key in merge_sig.parameters and key not in ("self", "items")
        }

    # -- Read ------------------------

    def _read_single_file(
        self,
        *args,  # noqa: ANN002
        **kwargs,
    ) -> T:
        """Execute ``cls.read()`` for a single file."""
        return self.registry.read(self._cls, *args, **kwargs)

    @staticmethod
    def _format_input_list(
        source: NamedReadable | list[NamedReadable],
    ) -> Sequence[NamedReadable]:
        """Format the input arguments to include a list of files."""
        # parse input as a list of files
        try:  # try and map to a list of file-like objects
            return file_list(source)
        except ValueError:  # otherwise treat as single file
            return [cast("NamedReadable", source)]

    def __call__(
        self,
        source: NamedReadable | list[NamedReadable],
        *args,  # noqa: ANN002
        format: str | None = None,  # noqa: A002
        cache: bool | None = None,
        parallel: int = 1,
        verbose: str | bool = False,
        merge_function: MergeProtocol | None = None,
        **kwargs,
    ) -> T:
        """Execute ``cls.read()``.

        This method generalises parallel reading of lists of files for
        any input class.


        This method also generalises downloading files from remote URLs
        over HTTP or via `fsspec`.
        """
        cls = self._cls

        # extract merge-specific kwargs by inspecting the merge method signature
        if merge_function is None:
            merge_function = self.merge
        merge_kwargs = self._merge_kwargs(kwargs, merge_function)

        # handle deprecated keyword
        if nproc := kwargs.pop("nproc", None):
            warnings.warn(
                f"the 'nproc' keyword to {cls.__name__}.read was renamed "
                "parallel; this warning will be an error in the future.",
                DeprecationWarning,
                stacklevel=2,
            )
            parallel = nproc

        # set default cache based on environment
        if cache is None:
            cache = bool_env("GWPY_CACHE", default=False)

        # get the input as a list of inputs
        sources = self._format_input_list(source)

        # handle progress based on the number of inputs
        show_progress: str | bool = False  # single file download progress
        if len(sources) == 1:
            show_progress = verbose
            verbose = False
        if verbose is True and format:
            verbose = f"Reading ({format})"
        elif verbose is True:
            verbose = "Reading"
        if verbose:
            progress = progress_bar(
                total=len(sources),
                desc=verbose,
            )
        else:
            progress = None

        remote_kwargs = {
            key: kwargs.pop(key) for key in (
                "sources",
                "http_headers",
                "use_fsspec",
                "fsspec_kwargs",
            ) if key in kwargs
        }

        # single file reader
        def _read(arg: NamedReadable) -> T:
            ctx: AbstractContextManager[NamedReadable | BinaryIO]
            # if arg is a str, presume it represents a URI, so try and open it
            if isinstance(arg, str):
                ctx = open_remote_file(
                    arg,
                    cache=cache,
                    show_progress=bool(show_progress),
                    **remote_kwargs,
                )
            # otherwise just pass it along unmodified
            else:
                ctx = nullcontext(arg)

            with ctx as file:
                return self.registry.read(
                    self._cls,
                    file,
                    *args,
                    format=format,
                    cache=cache,
                    **kwargs,
                )

        outputs: list[T] = []

        # read files in series
        if parallel <= 1:
            for src in sources:
                outputs.append(_read(src))
                if progress:
                    progress.update(1)
        # read all files in parallel threads
        else:
            with ThreadPoolExecutor(
                max_workers=parallel,
            ) as pool:
                futures = [pool.submit(_read, source) for source in sources]
                for future in as_completed(futures):
                    outputs.append(future.result())
                    if progress:
                        progress.update(1)

        # Merge results
        return merge_function(outputs, **merge_kwargs)


class UnifiedWrite(astropy_registry.UnifiedReadWrite, Generic[T]):
    """Base ``Class.write()`` implementation."""

    def __init__(
        self,
        instance: T,
        cls: type[T],
        registry: UnifiedIORegistry = default_registry,
    ) -> None:
        """Initialise a new `UnifiedWrite` instance."""
        super().__init__(
            instance,
            cls,
            "write",
            registry=registry,
        )

    def __call__(
        self,
        *args,  # noqa: ANN002
        **kwargs,
    ) -> None:
        """Execute ``instance.write()``."""
        instance = self._instance
        return self.registry.write(instance, *args, **kwargs)


# -- Fetch ---------------------------
# custom registry to support Klass.fetch

class UnifiedFetchRegistry(astropy_registry.UnifiedInputRegistry):
    """`UnifiedInputRegistry` hacked to support a ``.fetch()`` method.

    Fetch is a read-like operation that does not take a file or file-like
    object as input, but instead takes other arguments to fetch data from
    a single source, e.g. a database or web API.
    """

    name: Literal["read", "fetch", "get"] = "fetch"

    def __init__(self) -> None:
        """Initialise a new `UnifiedFetchRegistry` instance."""
        super().__init__()
        self._registries[self.name] = self._registries.pop("read")
        self._registries_order = (self.name, "identify")

    def _update__doc__(self, data_class: type, readwrite: str) -> None:
        """Update the docstring for ``data_class.<readwrite>()`` method."""
        if readwrite == "read":
            readwrite = self.name
        super()._update__doc__(data_class, readwrite)

        # replace the word 'format' with 'source' in the docstring
        class_readwrite_func = getattr(data_class, readwrite)
        doc = class_readwrite_func.__doc__
        if doc is None:
            return
        for search, replace in [
            # Leader for the sources table
            ("built-in formats", "built-in sources"),
            # Column heading in the sources table
            (r"\n(\s*)?Format", r"\n\1Source"),
        ]:
            doc = re.sub(search, replace, doc)
        class_readwrite_func.__class__.__doc__ = doc

    def _get_valid_format(
        self,
        mode: Literal["read", "fetch", "get"],
        cls: type,
        path: FileSystemPath | None,
        fileobj: FileLike | None,
        args: tuple,
        kwargs: dict,
    ) -> str:
        """Return the first valid format that can be used."""
        if mode.lower() == "read":
            mode = self.name
        return super()._get_valid_format(mode, cls, path, fileobj, args, kwargs)


fetch_registry = UnifiedFetchRegistry()


class UnifiedFetch(UnifiedRead[T], Generic[T]):
    """Base ``Class.fetch()`` implementation."""

    method: Literal["fetch", "get"] = "fetch"

    def merge(self, *args, **kwargs) -> Never:
        """Fetch does not support merging."""
        msg = f"{self._cls.__name__}.{self.method} does not support merging"
        raise NotImplementedError(msg)

    def __init__(
        self,
        instance: T,
        cls: type[T],
        registry: UnifiedFetchRegistry = fetch_registry,
    ) -> None:
        """Initialise a new `UnifiedFetch` instance."""
        super().__init__(
            instance,
            cls,
            registry=registry,
        )

    def __call__(  # type: ignore[override]
        self,
        *args,  # noqa: ANN002
        source: str | None = None,
        **kwargs,
    ) -> T:
        """Execute ``cls.fetch()``."""
        # 'read' using the registered format.
        return self.registry.read(
            self._cls,
            *args,
            format=source,
            **kwargs,
        )

    def help(
        self,
        source: str | None = None,
        out: FileLike | None = None,
    ) -> None:
        """Output help documentation for the specified `UnifiedFetch` ``source``.

        By default the help output is printed to the console via ``pydoc.pager``.
        Instead one can supplied a file handle object as ``out`` and the output
        will be written to that handle.

        Parameters
        ----------
        source : str
            `UnifiedFetch` source (format) name, e.g. 'sql' or 'gwosc'.

        out : None or file-like
            Output destination (default is stdout via a pager)
        """
        cls = self._cls

        # Get reader or writer function associated with the registry
        get_func = self._registry.get_reader
        try:
            if source:
                read_write_func = get_func(source, cls)
        except astropy_registry.IORegistryError as err:
            reader_doc = "ERROR: " + str(err)
        else:
            if source:
                # Format-specific
                header = (
                    f"{cls.__name__}.{self.method}(source='{source}') documentation\n"
                )
                doc = read_write_func.__doc__
            else:
                # General docs
                header = f"{cls.__name__}.{self.method} general documentation\n"
                doc = getattr(cls, self.method).__doc__

            reader_doc = re.sub(".", "=", header)
            reader_doc += header
            reader_doc += re.sub(".", "=", header)
            reader_doc += os.linesep
            if doc is not None:
                reader_doc += inspect.cleandoc(doc)

        if out is None:
            pydoc.pager(reader_doc)
        else:
            out.write(reader_doc)


# -- Get -----------------------------
# custom registry to support Klass.get

class UnifiedGetRegistry(UnifiedFetchRegistry):
    """Unified I/O registry for providing a multi-source ``.get()`` method.

    Each registered reader should work with the target class as normal.

    Each registered identified should return an iterable (generator) of
    keyword argument sets that should be tried, in order, to provide the
    necessary data.

    See `gwpy.timeseries.io.nds2` for a working example.
    """

    name = "get"
    _identifiers: dict[tuple[str, type], Callable[..., Sequence[dict[str, Any]]]]

    def identify_sources(
        self,
        source: str | Iterable[str | dict[str, Any]] | None,
        data_class_required: type,
        args: list[Any],
        kwargs: dict,
    ) -> list[tuple[str, dict[str, Any]]]:
        """Identify all valid sources for the given arguments.

        Unlike the standard registry `identify_format` method, this returns
        a `list` of `dict`, each of which should be tried in turn with the
        keyword arguments applied for each source.

        This allows the `Klass.get` method to try multiple different sources
        to get some data.
        """
        # format input
        if isinstance(source, str):
            source = [source]
        sources: list[dict[str, Any]] = []
        source_names: set[str] = set()
        for src in source or []:
            if isinstance(src, str):
                src = {"source": src}  # noqa: PLW2901
            if not isinstance(src, dict):
                msg = "source must be a string or a dictionary of keyword arguments"
                raise TypeError(msg)
            try:
                source_names.add(src["source"])
            except KeyError as exc:
                msg = "source dictionary must contain a 'source' key"
                raise ValueError(msg) from exc
            sources.append(src)

        out: list[tuple[str, dict[str, Any]]] = []

        def _match_source(
            source: str | None = None,
            **source_kw,
        ) -> Iterable[tuple[str, Any]]:
            """Return all matching sources for the given source name and keywords.

            Each source identifier function returns a list of keyword argument dicts
            that should be tried in order.
            If the ``priority`` key is present in the dict, the results are
            sorted by priority across all registered sources
            (lowest first, default is 10).
            """
            match: list[tuple[str, dict[str, Any]]] = []
            for data_source, data_class in self._identifiers:
                # If the user specified a source, and it wasn't this one, skip it
                if source is not None and data_source != source:
                    continue
                # Exclude identifiers for a parent class where the child class
                # has its own registration _for this source type_.
                if not self._is_best_match(
                    data_class_required,
                    data_class,
                    filter(lambda klass: klass[0] == data_source, self._identifiers),
                ):
                    continue
                if new := self._identifiers[(data_source, data_class)](
                    self.name,
                    *args,
                    **kwargs,
                    **source_kw,
                ):
                    match.extend((data_source, kw) for kw in new)
            match.sort(key=lambda x: x[1].pop("priority", 10))  # sort by priority
            return match

        if not sources:
            sources = [{"source": None}]

        for source in sources:
            out.extend(_match_source(**source))

        return out


get_registry = UnifiedGetRegistry()


class UnifiedGet(UnifiedFetch, Generic[T]):
    """Unified I/O ``.get()`` implementation.

    This is similar to `UnifiedRead` or `UnifiedFetch` except that the getter
    will iterate over multiple sources returned by the
    `UnifiedGetRegistry.identify_sources` method, to retrieve the data any
    way it can.
    """

    method = "get"

    def __init__(
        self,
        instance: T,
        cls: type[T],
        registry: UnifiedGetRegistry = get_registry,
        module: str | None = None,
    ) -> None:
        """Initialise a new `UnifiedFetch` instance."""
        super().__init__(
            instance,
            cls,
            registry=registry,
        )
        self.logger = logging.getLogger(module or cls.__module__)

    def __call__(  # type: ignore[override]
        self,
        *args,
        source: str | list[str | dict[str, Any]] | None = None,
        **kwargs,
    ) -> T:
        """Execute ``cls.get()``."""
        reg = self.registry
        sources = reg.identify_sources(
            source,
            self._cls,
            args,
            kwargs,
        )
        nsources = len(sources)
        if not nsources:
            msg = "no valid sources found"
            raise ValueError(msg)
        self.logger.info("Found %d possible sources", nsources)

        errors = []

        # loop over each source
        for i, (src, source_kw) in enumerate(sources, start=1):
            self.logger.info("Attemping access with '%s' [%d/%d]", src, i, nsources)
            self.logger.debug(
                "Using options: %s",
                {k: v for k, v in source_kw.items() if v is not None},
            )
            try:
                return self._get(src, source_kw, *args, **kwargs)
            except Exception as exc:
                exc.add_note(f"Error getting data from {src} [{i}/{nsources}]")
                errors.append(exc)
                self.logger.debug(
                    "Failed to get data with %s: %s: %s",
                    src,
                    type(exc).__name__,
                    str(exc),
                )

        msg = "failed to get data from any source"
        raise GetExceptionGroup(msg, errors)

    def _get(
        self,
        source: str,
        source_kw: dict[str, Any],
        *args,
        **kwargs,
    ) -> T:
        """Try and get data from a single source."""
        getter = self.registry.get_reader(source, self._cls)

        # parse arguments
        sig = inspect.signature(getter)

        # Check if getter accepts **kwargs (VAR_KEYWORD parameter)
        has_var_keyword = any(
            p.kind == inspect.Parameter.VAR_KEYWORD
            for p in sig.parameters.values()
        )

        if has_var_keyword:
            # Pass all kwargs except framework parameters handled by __call__
            framework_params = {"source"}
            source_kw |= {
                key: val for key, val in kwargs.items()
                if key not in framework_params and val is not None
            }
        else:
            # Only pass kwargs that match explicitly named parameters
            params = [
                p.name
                for p in sig.parameters.values()
                if p.kind in (p.POSITIONAL_OR_KEYWORD, p.KEYWORD_ONLY)
            ]
            source_kw |= {
                key: val for key, val in kwargs.items()
                if key in params and val is not None
            }

        return getter(
            *args,
            **source_kw,
        )


# -- utilities -----------------------

def inherit_unified_io(klass: type) -> type:
    """Re-register all Unified I/O readers/writers/identifiers from a parent to a child.

    Only works with the first parent in the inheritance tree.

    This allows the Unified I/O registrations for the child class to be
    modified independently of the parent.
    """
    parent = klass.__mro__[1]
    parent_registry = parent.read.registry  # type: ignore[attr-defined]
    child_registry = klass.read.registry  # type: ignore[attr-defined]
    for row in parent_registry.get_formats(data_class=parent):
        name = row["Format"]

        # read
        if row["Read"].lower() == "yes":
            child_registry.register_reader(
                name,
                klass,
                parent_registry.get_reader(name, parent),
                force=False,
            )
        # write
        if row["Write"].lower() == "yes":
            child_registry.register_writer(
                name,
                klass,
                parent_registry.get_writer(name, parent),
                force=False,
            )
        # identify
        if row["Auto-identify"].lower() == "yes":
            child_registry.register_identifier(
                name,
                klass,
                parent_registry._identifiers[(name, parent)],
                force=False,
            )
    return klass



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Input/output routines for gravitational-wave frame (GWF) format files.

The API for GWF I/O integration is defined by the core module, which
specifies that each backend module provides two functions:

`read`
    A function that takes in a source reference (file path) and a list of
    channels to read, and returns a `TimeSeriesDict` of data.

`write`
    A function that takes in a `TimeSeriesDict` of data and a target reference
    (file path) and writes the data to that file.

See `gwpy.timeseries.io.gwf.core.read_timeseriesdict` and
`gwpy.timeseries.io.gwf.core.write_timeseriesdict` for full function
signatures and docstrings.

A backend module can then register itself to act as a backend for the
`format='gwf'` series reader by calling the
`gwpy.timeseries.io.gwf.core.register_gwf_backend` function.
"""

# basic routines
from . import core

# backends
for backend_mod_name in (
    "frameCPP",
    "framel",
    "lalframe",
):
    core.register_gwf_backend(backend_mod_name)

BACKENDS = core.BACKENDS



# Copyright (c) 2019-2020 Patrick Godwin
#               2020 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Read events from SNAX."""

from __future__ import annotations

import warnings
from typing import TYPE_CHECKING

from astropy.io.misc.hdf5 import read_table_hdf5
from astropy.table import vstack

from ...io.hdf5 import with_read_hdf5
from .. import EventTable
from .utils import (
    read_with_columns,
    read_with_where,
)

if TYPE_CHECKING:
    from collections.abc import Iterable
    from typing import Literal

    import h5py
    from astropy.table import Table

__author__ = "Patrick Godwin <patrick.godwin@ligo.org>"


@read_with_columns
@read_with_where
@with_read_hdf5
def table_from_file(
    source: h5py.Group,
    channels: Iterable[str] | None = None,
    on_missing: Literal["error", "warn", "ignore"] = "error",
    *,
    compact: bool = False,
) -> Table:
    """Read an `EventTable` from a SNAX HDF5 file.

    Parameters
    ----------
    source : `str`, `pathlib.Path`, `file`, `h5py.File`
        The file path of open `h5py` object from which to read the data.

    channels : `str` or `list`, optional
        The channel(s) to read data from. if no channels selected,
        read in all channels.

    on_missing: `str`, optional
        How to proceed when channels requested are missing. One of:

        "warn"
            Emit a warning when missing channels are discovered.

        "error"
            Raise an exception.

        Default is ``"error"``

    compact : `bool`, optional
        Whether to store a compact integer representation in the channel
        column rather than the full channel name, instead storing a mapping
        (`channel_map`) in the table metadata.
        Default is `False`.

    Returns
    -------
    table : `~astropy.table.Table`
    """
    # format channels appropriately
    if isinstance(channels, str):
        channels = [channels]

    # only query channels contained in file
    # if channels not specified, load all of them
    if channels is None:
        channels = source.keys()
    else:
        channels = set(channels)
        found = set(source.keys())
        missing = channels - found
        # check whether missing channels should
        # be an error or simply a warning
        if missing and on_missing != "ignore":
            msg = "requested channels not found in SNAX file: '{}'".format(
                "', '".join(missing),
            )
            if on_missing == "error":
                raise ValueError(msg)
            if on_missing == "warn":
                warnings.warn(msg, stacklevel=2)
            else:
                msg = "on_missing argument must be one of 'error', 'warn', or 'ignore'"
                raise ValueError(msg)
        channels = channels & found

    # read data, adding in 'channel' column
    # to preserve uniqueness across channels
    tables = []
    for channel in channels:
        table = vstack(
            list(map(read_table_hdf5, source[channel].values())),
            join_type="exact",
            metadata_conflicts="error",
        )
        # determine whether to store a compact
        # representation of the channel, storing
        # the mapping to table metadata instead
        if compact:
            table["channel"] = hash(channel)
            table.meta["channel_map"] = {hash(channel): channel}
        else:
            table["channel"] = channel
        tables.append(table)

    # combine results
    return vstack(tables, join_type="exact", metadata_conflicts="error")


# register for unified I/O
EventTable.read.registry.register_reader(
    "hdf5.snax",
    EventTable,
    table_from_file,
)
